import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from .LilyPondGrob import LilyPondGrob


class LilyPondEngraver(AbjadValueObject):
    """
    LilyPond engraver.

    ..  container:: example

        >>> engraver = abjad.LilyPondEngraver('Auto_beam_engraver')
        >>> print(format(engraver))
        abjad.LilyPondEngraver(
            name='Auto_beam_engraver',
            )

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_name',
        )

    _identity_map: typing.Dict[str, 'LilyPondEngraver'] = {}

    ### CONSTRUCTOR ###

    def __new__(class_, name='Note_heads_engraver'):
        if name in class_._identity_map:
            obj = class_._identity_map[name]
        else:
            obj = object.__new__(class_)
            class_._identity_map[name] = obj
        return obj

    ### INITIALIZER ###

    def __init__(self, name: str = 'Note_heads_engraver') -> None:
        from abjad.ly import engravers
        assert name in engravers
        self._name = name

    ### PUBLIC METHODS ###

    @staticmethod
    def list_all_engravers() -> typing.Tuple['LilyPondEngraver', ...]:
        """
        Lists all engravers.

        ..  container:: example

            >>> for lilypond_engraver in abjad.LilyPondEngraver.list_all_engravers():
            ...     lilypond_engraver
            ...
            LilyPondEngraver(name='Accidental_engraver')
            LilyPondEngraver(name='Ambitus_engraver')
            LilyPondEngraver(name='Arpeggio_engraver')
            LilyPondEngraver(name='Auto_beam_engraver')
            LilyPondEngraver(name='Axis_group_engraver')
            LilyPondEngraver(name='Balloon_engraver')
            LilyPondEngraver(name='Bar_engraver')
            LilyPondEngraver(name='Bar_number_engraver')
            LilyPondEngraver(name='Beam_collision_engraver')
            LilyPondEngraver(name='Beam_engraver')
            LilyPondEngraver(name='Beam_performer')
            LilyPondEngraver(name='Bend_engraver')
            LilyPondEngraver(name='Break_align_engraver')
            LilyPondEngraver(name='Breathing_sign_engraver')
            LilyPondEngraver(name='Chord_name_engraver')
            LilyPondEngraver(name='Chord_tremolo_engraver')
            LilyPondEngraver(name='Clef_engraver')
            LilyPondEngraver(name='Cluster_spanner_engraver')
            LilyPondEngraver(name='Collision_engraver')
            LilyPondEngraver(name='Completion_heads_engraver')
            LilyPondEngraver(name='Completion_rest_engraver')
            LilyPondEngraver(name='Concurrent_hairpin_engraver')
            LilyPondEngraver(name='Control_track_performer')
            LilyPondEngraver(name='Cue_clef_engraver')
            LilyPondEngraver(name='Custos_engraver')
            LilyPondEngraver(name='Default_bar_line_engraver')
            LilyPondEngraver(name='Dot_column_engraver')
            LilyPondEngraver(name='Dots_engraver')
            LilyPondEngraver(name='Double_percent_repeat_engraver')
            LilyPondEngraver(name='Drum_note_performer')
            LilyPondEngraver(name='Drum_notes_engraver')
            LilyPondEngraver(name='Dynamic_align_engraver')
            LilyPondEngraver(name='Dynamic_engraver')
            LilyPondEngraver(name='Dynamic_performer')
            LilyPondEngraver(name='Engraver')
            LilyPondEngraver(name='Episema_engraver')
            LilyPondEngraver(name='Extender_engraver')
            LilyPondEngraver(name='Figured_bass_engraver')
            LilyPondEngraver(name='Figured_bass_position_engraver')
            LilyPondEngraver(name='Fingering_column_engraver')
            LilyPondEngraver(name='Fingering_engraver')
            LilyPondEngraver(name='Font_size_engraver')
            LilyPondEngraver(name='Footnote_engraver')
            LilyPondEngraver(name='Forbid_line_break_engraver')
            LilyPondEngraver(name='Fretboard_engraver')
            LilyPondEngraver(name='Glissando_engraver')
            LilyPondEngraver(name='Grace_auto_beam_engraver')
            LilyPondEngraver(name='Grace_beam_engraver')
            LilyPondEngraver(name='Grace_engraver')
            LilyPondEngraver(name='Grace_spacing_engraver')
            LilyPondEngraver(name='Grid_line_span_engraver')
            LilyPondEngraver(name='Grid_point_engraver')
            LilyPondEngraver(name='Grob_pq_engraver')
            LilyPondEngraver(name='Horizontal_bracket_engraver')
            LilyPondEngraver(name='Hyphen_engraver')
            LilyPondEngraver(name='Instrument_name_engraver')
            LilyPondEngraver(name='Instrument_switch_engraver')
            LilyPondEngraver(name='Keep_alive_together_engraver')
            LilyPondEngraver(name='Key_engraver')
            LilyPondEngraver(name='Key_performer')
            LilyPondEngraver(name='Kievan_ligature_engraver')
            LilyPondEngraver(name='Laissez_vibrer_engraver')
            LilyPondEngraver(name='Ledger_line_engraver')
            LilyPondEngraver(name='Ligature_bracket_engraver')
            LilyPondEngraver(name='Lyric_engraver')
            LilyPondEngraver(name='Lyric_performer')
            LilyPondEngraver(name='Mark_engraver')
            LilyPondEngraver(name='Measure_grouping_engraver')
            LilyPondEngraver(name='Melody_engraver')
            LilyPondEngraver(name='Mensural_ligature_engraver')
            LilyPondEngraver(name='Metronome_mark_engraver')
            LilyPondEngraver(name='Midi_control_function_performer')
            LilyPondEngraver(name='Multi_measure_rest_engraver')
            LilyPondEngraver(name='New_fingering_engraver')
            LilyPondEngraver(name='Note_head_line_engraver')
            LilyPondEngraver(name='Note_heads_engraver')
            LilyPondEngraver(name='Note_name_engraver')
            LilyPondEngraver(name='Note_performer')
            LilyPondEngraver(name='Note_spacing_engraver')
            LilyPondEngraver(name='Ottava_spanner_engraver')
            LilyPondEngraver(name='Output_property_engraver')
            LilyPondEngraver(name='Page_turn_engraver')
            LilyPondEngraver(name='Paper_column_engraver')
            LilyPondEngraver(name='Parenthesis_engraver')
            LilyPondEngraver(name='Part_combine_engraver')
            LilyPondEngraver(name='Percent_repeat_engraver')
            LilyPondEngraver(name='Phrasing_slur_engraver')
            LilyPondEngraver(name='Piano_pedal_align_engraver')
            LilyPondEngraver(name='Piano_pedal_engraver')
            LilyPondEngraver(name='Piano_pedal_performer')
            LilyPondEngraver(name='Pitch_squash_engraver')
            LilyPondEngraver(name='Pitched_trill_engraver')
            LilyPondEngraver(name='Pure_from_neighbor_engraver')
            LilyPondEngraver(name='Repeat_acknowledge_engraver')
            LilyPondEngraver(name='Repeat_tie_engraver')
            LilyPondEngraver(name='Rest_collision_engraver')
            LilyPondEngraver(name='Rest_engraver')
            LilyPondEngraver(name='Rhythmic_column_engraver')
            LilyPondEngraver(name='Scheme_engraver')
            LilyPondEngraver(name='Script_column_engraver')
            LilyPondEngraver(name='Script_engraver')
            LilyPondEngraver(name='Script_row_engraver')
            LilyPondEngraver(name='Separating_line_group_engraver')
            LilyPondEngraver(name='Slash_repeat_engraver')
            LilyPondEngraver(name='Slur_engraver')
            LilyPondEngraver(name='Slur_performer')
            LilyPondEngraver(name='Spacing_engraver')
            LilyPondEngraver(name='Span_arpeggio_engraver')
            LilyPondEngraver(name='Span_bar_engraver')
            LilyPondEngraver(name='Span_bar_stub_engraver')
            LilyPondEngraver(name='Spanner_break_forbid_engraver')
            LilyPondEngraver(name='Staff_collecting_engraver')
            LilyPondEngraver(name='Staff_performer')
            LilyPondEngraver(name='Staff_symbol_engraver')
            LilyPondEngraver(name='Stanza_number_align_engraver')
            LilyPondEngraver(name='Stanza_number_engraver')
            LilyPondEngraver(name='Stem_engraver')
            LilyPondEngraver(name='System_start_delimiter_engraver')
            LilyPondEngraver(name='Tab_note_heads_engraver')
            LilyPondEngraver(name='Tab_staff_symbol_engraver')
            LilyPondEngraver(name='Tab_tie_follow_engraver')
            LilyPondEngraver(name='Tempo_performer')
            LilyPondEngraver(name='Text_engraver')
            LilyPondEngraver(name='Text_spanner_engraver')
            LilyPondEngraver(name='Tie_engraver')
            LilyPondEngraver(name='Tie_performer')
            LilyPondEngraver(name='Time_signature_engraver')
            LilyPondEngraver(name='Time_signature_performer')
            LilyPondEngraver(name='Timing_translator')
            LilyPondEngraver(name='Translator')
            LilyPondEngraver(name='Trill_spanner_engraver')
            LilyPondEngraver(name='Tuplet_engraver')
            LilyPondEngraver(name='Tweak_engraver')
            LilyPondEngraver(name='Vaticana_ligature_engraver')
            LilyPondEngraver(name='Vertical_align_engraver')
            LilyPondEngraver(name='Volta_engraver')

        """
        from abjad.ly import engravers
        return tuple(LilyPondEngraver(name=name) for name in sorted(engravers))

    ### PUBLIC PROPERTIES ###

    @property
    def grobs(self) -> typing.Tuple[LilyPondGrob, ...]:
        """
        Gets LilyPond engraver's created grobs.

        ..  container:: example

            >>> engraver = abjad.LilyPondEngraver('Auto_beam_engraver')
            >>> for grob in engraver.grobs:
            ...     grob
            ...
            LilyPondGrob(name='Beam')

        """
        from abjad.ly import engravers
        dictionary = engravers[self.name]
        assert isinstance(dictionary, dict), repr(dictionary)
        return tuple(
            LilyPondGrob(name=name)
            for name in dictionary['grobs_created']
            )

    @property
    def name(self) -> str:
        """
        Gets name of LilyPond engraver.

        ..  container:: example

            >>> engraver = abjad.LilyPondEngraver('Auto_beam_engraver')
            >>> engraver.name
            'Auto_beam_engraver'

        """
        return self._name

    @property
    def property_names(self) -> typing.Tuple[str, ...]:
        """
        Gets LilyPond engraver's property names.

        ..  container:: example

            >>> engraver = abjad.LilyPondEngraver('Auto_beam_engraver')
            >>> for property_name in engraver.property_names:
            ...     property_name
            ...
            'autoBeaming'
            'baseMoment'
            'beamExceptions'
            'beamHalfMeasure'
            'beatStructure'
            'subdivideBeams'

        """
        from abjad.ly import engravers
        dictionary = engravers[self.name]
        assert isinstance(dictionary, dict), repr(dictionary)
        property_names: typing.Set[str] = set()
        property_names.update(dictionary['properties_read'])
        property_names.update(dictionary['properties_written'])
        return tuple(sorted(property_names))



import os
import re
import shutil
import tempfile
from abjad.system.AbjadObject import AbjadObject


class PersistenceManager(AbjadObject):
    """
    Persistence manager.

    ..  container:: example

        >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
        >>> abjad.show(staff) # doctest: +SKIP

        >>> abjad.persist(staff)
        PersistenceManager(client=Staff("c'4 e'4 d'4 f'4"))

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_client',
        )

    _png_page_pattern = re.compile(r'.+page(\d+)\.png')

    ### INITIALIZER ###

    def __init__(self, client=None):
        self._client = client

    ### PUBLIC METHODS ###

    def as_ly(
        self,
        ly_file_path=None,
        illustrate_function=None,
        strict=None,
        **keywords
        ):
        """
        Persists client as LilyPond file.

        Autogenerates file path when ``ly_file_path`` is none.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
            >>> for x in persist(staff).as_ly('~/example.ly'): # doctest: +SKIP
            ...     x
            ...
            '/Users/josiah/Desktop/example.ly'
            0.04491996765136719

        Returns output path and elapsed formatting time when LilyPond output is
        written.
        """
        import abjad
        if strict is not None:
            assert isinstance(strict, int), repr(strict)
        if illustrate_function is None:
            assert hasattr(self._client, '__illustrate__')
            illustrate_function = self._client.__illustrate__
        lilypond_file = illustrate_function(**keywords)
        if ly_file_path is None:
            ly_file_name = abjad.IOManager.get_next_output_file_name()
            ly_file_path = os.path.join(
                abjad.abjad_configuration.abjad_output_directory,
                ly_file_name,
                )
        else:
            ly_file_path = str(ly_file_path)
            ly_file_path = os.path.expanduser(ly_file_path)
        assert ly_file_path.endswith('.ly'), ly_file_path
        timer = abjad.Timer()
        with timer:
            string = lilypond_file.__format__(
                format_specification='lilypond',
                )
            if isinstance(strict, int):
                string = abjad.LilyPondFormatManager.align_tags(string, strict)
        abjad_formatting_time = timer.elapsed_time
        directory = os.path.dirname(ly_file_path)
        abjad.IOManager._ensure_directory_existence(directory)
        with open(ly_file_path, 'w') as file_pointer:
            file_pointer.write(string)
        return ly_file_path, abjad_formatting_time

    def as_midi(self, midi_file_path=None, remove_ly=False, **keywords):
        """
        Persists client as MIDI file.

        Autogenerates file path when ``midi_file_path`` is none.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
            >>> for x in persist(staff).as_midi(): # doctest: +SKIP
            ...     x
            ...
            '/Users/josiah/.abjad/output/1415.midi'
            0.07831692695617676
            1.0882699489593506

        Returns output path, elapsed formatting time and elapsed rendering
        time.
        """
        import abjad
        assert hasattr(self._client, '__illustrate__')
        illustration = self._client.__illustrate__(**keywords)
        assert hasattr(illustration, 'score_block')
        block = abjad.Block(name='midi')
        illustration.score_block.items.append(block)
        if midi_file_path is not None:
            midi_file_path = os.path.expanduser(midi_file_path)
            without_extension = os.path.splitext(midi_file_path)[0]
            ly_file_path = '{}.ly'.format(without_extension)
        else:
            ly_file_path = None
        result = type(self)(illustration).as_ly(ly_file_path, **keywords)
        ly_file_path, abjad_formatting_time = result
        timer = abjad.Timer()
        with timer:
            success = abjad.IOManager.run_lilypond(ly_file_path)
        lilypond_rendering_time = timer.elapsed_time
        if os.name == 'nt':
            extension = 'mid'
        else:
            extension = 'midi'
        midi_file_path = '{}.{}'.format(
            os.path.splitext(ly_file_path)[0],
            extension,
            )
        if remove_ly:
            os.remove(ly_file_path)
        return (
            midi_file_path,
            abjad_formatting_time,
            lilypond_rendering_time,
            success,
            )

    def as_module(self, module_file_path, object_name):
        """
        Persists client as Python module.

        ..  container:: example

            >>> timespans = abjad.TimespanList([
            ...     abjad.Timespan(0, 1),
            ...     abjad.Timespan(2, 4),
            ...     abjad.Timespan(6, 8),
            ...     ])
            >>> abjad.persist(timespans).as_module( # doctest: +SKIP
            ...     '~/example.py',
            ...     'timespans',
            ...     )

        Returns none.
        """
        import abjad
        agent = abjad.StorageFormatManager(self._client)
        result = []
        import_statements = agent.get_import_statements()
        result.extend(import_statements)
        result.extend(('', ''))
        if (
            '_get_storage_format_specification' in dir(self._client) or
            '_get_format_specification' in dir(self._client)
            ):
            storage_pieces = format(self._client, 'storage')
        else:
            try:
                storage_pieces = agent._dispatch_formatting(self._client)
                storage_pieces = ''.join(storage_pieces)
            except ValueError:
                storage_pieces = repr(self._client)
        storage_pieces = storage_pieces.splitlines()
        line = '{} = {}'.format(object_name, storage_pieces[0])
        result.append(line)
        result.extend(storage_pieces[1:])
        result = '\n'.join(result)
        module_file_path = os.path.expanduser(module_file_path)
        directory = os.path.dirname(module_file_path)
        abjad.IOManager._ensure_directory_existence(directory)
        with open(module_file_path, 'w') as f:
            f.write(result)

    def as_pdf(
        self,
        pdf_file_path=None,
        illustrate_function=None,
        remove_ly=False,
        strict=None,
        **keywords
        ):
        """
        Persists client as PDF.

        Autogenerates file path when ``pdf_file_path`` is none.

        ..  container:: example

            >>> staff = abjad.Staff("c'4 e'4 d'4 f'4")
            >>> for x in persist(staff).as_pdf(): # doctest: +SKIP
            ...     x
            ...
            '/Users/josiah/.abjad/output/1416.pdf'
            0.047142982482910156
            0.7839350700378418

        Returns output path, elapsed formatting time and elapsed rendering
        time when PDF output is written.
        """
        import abjad
        if strict is not None:
            assert isinstance(strict, int), repr(strict)
        if illustrate_function is None:
            assert hasattr(self._client, '__illustrate__'), repr(self._client)
        if pdf_file_path is not None:
            pdf_file_path = str(pdf_file_path)
            pdf_file_path = os.path.expanduser(pdf_file_path)
            without_extension = os.path.splitext(pdf_file_path)[0]
            ly_file_path = '{}.ly'.format(without_extension)
        else:
            ly_file_path = None
        result = self.as_ly(
            ly_file_path,
            illustrate_function=illustrate_function,
            strict=strict,
            **keywords
            )
        ly_file_path, abjad_formatting_time = result
        without_extension = os.path.splitext(ly_file_path)[0]
        pdf_file_path = '{}.pdf'.format(without_extension)
        timer = abjad.Timer()
        with timer:
            success = abjad.IOManager.run_lilypond(ly_file_path)
        lilypond_rendering_time = timer.elapsed_time
        if remove_ly:
            os.remove(ly_file_path)
        return (
            pdf_file_path,
            abjad_formatting_time,
            lilypond_rendering_time,
            success,
            )

    def as_png(
        self,
        png_file_path=None,
        remove_ly=False,
        illustrate_function=None,
        **keywords
        ):
        """
        Persists client as PNG.

        ..  container:: example

            >>> staff = abjad.Staff()
            >>> measure = abjad.Measure((4, 4), "c'4 d'4 e'4 f'4")
            >>> command = abjad.LilyPondLiteral(r'\break', 'after')
            >>> abjad.attach(command, measure[-1])
            >>> staff.extend(measure * 200)

            >>> result = persist(staff).as_png() # doctest: +SKIP
            >>> for x in result[0]: # doctest: +SKIP
            ...     x
            ...
            '/Users/josiah/Desktop/test-page1.png'
            '/Users/josiah/Desktop/test-page2.png'
            '/Users/josiah/Desktop/test-page3.png'
            '/Users/josiah/Desktop/test-page4.png'
            '/Users/josiah/Desktop/test-page5.png'
            '/Users/josiah/Desktop/test-page6.png'
            '/Users/josiah/Desktop/test-page7.png'
            '/Users/josiah/Desktop/test-page8.png'
            '/Users/josiah/Desktop/test-page9.png'
            '/Users/josiah/Desktop/test-page10.png'
            '/Users/josiah/Desktop/test-page11.png'
            '/Users/josiah/Desktop/test-page12.png'
            '/Users/josiah/Desktop/test-page13.png'
            '/Users/josiah/Desktop/test-page14.png'
            '/Users/josiah/Desktop/test-page15.png'

        Autogenerates file path when ``png_file_path`` is none.

        Returns output path(s), elapsed formatting time and elapsed rendering
        time.
        """
        import abjad
        if illustrate_function is None:
            assert hasattr(self._client, '__illustrate__')
        if png_file_path is not None:
            png_file_path = os.path.expanduser(png_file_path)
            without_extension = os.path.splitext(png_file_path)[0]
            ly_file_path = '{}.ly'.format(without_extension)
        else:
            ly_file_path = None
        result = self.as_ly(
            ly_file_path,
            illustrate_function=illustrate_function,
            **keywords
            )

        ly_file_path, abjad_formatting_time = result

        original_directory = os.path.split(ly_file_path)[0]
        original_ly_file_path = ly_file_path
        temporary_directory = tempfile.mkdtemp()
        temporary_ly_file_path = os.path.join(
            temporary_directory,
            os.path.split(ly_file_path)[1],
            )
        shutil.copy(original_ly_file_path, temporary_ly_file_path)

        timer = abjad.Timer()
        with timer:
            success = abjad.IOManager.run_lilypond(
                temporary_ly_file_path,
                flags='--png',
                )
        lilypond_rendering_time = timer.elapsed_time

        png_file_paths = []
        for file_name in os.listdir(temporary_directory):
            if not file_name.endswith('.png'):
                continue
            source_png_file_path = os.path.join(
                temporary_directory,
                file_name,
                )
            target_png_file_path = os.path.join(
                original_directory,
                file_name,
                )
            shutil.move(source_png_file_path, target_png_file_path)
            png_file_paths.append(target_png_file_path)
        shutil.rmtree(temporary_directory)

        if remove_ly:
            os.remove(ly_file_path)

        if 1 < len(png_file_paths):
            png_file_paths.sort(
                key=lambda x: int(self._png_page_pattern.match(x).groups()[0]),
                )

        return (
            tuple(png_file_paths),
            abjad_formatting_time,
            lilypond_rendering_time,
            success,
            )

    ### PRIVATE PROPERTIES ###

    @property
    def client(self):
        """
        Client of persistence manager.

        Returns component or selection.
        """
        return self._client



import time
from abjad.system.AbjadObject import AbjadObject


class DateTimeToken(AbjadObject):
    """
    A LilyPond file date / time token.

    ..  container:: example

        >>> abjad.DateTimeToken()
        DateTimeToken()

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_date_string',
        )

    ### INITIALIZER ###

    def __init__(self, date_string=None):
        assert isinstance(date_string, (str, type(None)))
        self._date_string = date_string

    ### SPECIAL METHODS ###

    def __format__(self, format_specification=''):
        """
        Formats date / time token.

        ..  container:: example

            >>> token = abjad.DateTimeToken()
            >>> print(format(token)) # doctest: +SKIP
            2014-01-04 14:42

        Returns string.
        """
        import abjad
        if format_specification in ('', 'lilypond'):
            return self._get_lilypond_format()
        elif format_specification == 'storage':
            return abjad.StorageFormatManager(self).get_storage_format()
        return str(self)

    def __repr__(self):
        """
        Gets interpreter representation of date / time token.

        ..  container:: example

            >>> abjad.DateTimeToken()
            DateTimeToken()

        Returns string.
        """
        date_string = self._date_string or ''
        return '{}({})'.format(type(self).__name__, date_string)

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return self.date_string

    ### PUBLIC PROPERTIES ###

    @property
    def date_string(self):
        """
        Gets date string of date / time token.

        ..  container:: example

            >>> token = abjad.DateTimeToken()
            >>> token.date_string # doctest: +SKIP
            '2014-01-23 12:21'

        Returns string.
        """
        date_string = self._date_string or time.strftime('%Y-%m-%d %H:%M')
        return date_string



import abc
import functools
import math
import numbers
from abjad import mathtools
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.StorageFormatManager import StorageFormatManager
from . import constants


@functools.total_ordering
class Pitch(AbjadValueObject):
    """
    Abstract pitch.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_pitch_class',
        '_octave',
        )

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self, argument, accidental=None, arrow=None, octave=None):
        import abjad
        if isinstance(argument, str):
            match = constants._comprehensive_pitch_name_regex.match(argument)
            if not match:
                match = constants._comprehensive_pitch_class_name_regex.match(argument)
            if not match:
                message = 'can not instantiate {} from {!r}.'
                message = message.format(type(self).__name__, argument)
                raise ValueError(message)
            group_dict = match.groupdict()
            _dpc_name = group_dict['diatonic_pc_name'].lower()
            _dpc_number = constants._diatonic_pc_name_to_diatonic_pc_number[_dpc_name]
            _alteration = abjad.Accidental(group_dict['comprehensive_accidental']).semitones
            _octave = abjad.Octave(group_dict.get('comprehensive_octave', '')).number
            self._from_named_parts(_dpc_number, _alteration, _octave)
        elif isinstance(argument, numbers.Number):
            self._from_number(argument)
        elif isinstance(argument, (abjad.Pitch, abjad.PitchClass)):
            self._from_pitch_or_pitch_class(argument)
        elif isinstance(argument, tuple) and len(argument) == 2:
            _pitch_class = abjad.NamedPitchClass(argument[0])
            _octave = abjad.Octave(argument[1])
            self._from_named_parts(
                _pitch_class._get_diatonic_pc_number(),
                _pitch_class._get_alteration(),
                _octave.number,
            )
        elif hasattr(argument, 'written_pitch'):
            self._from_pitch_or_pitch_class(argument.written_pitch)
        elif isinstance(argument, abjad.Chord) and len(argument.note_heads):
            self._from_pitch_or_pitch_class(argument.note_heads[0])
        else:
            message = 'can not instantiate {} from {!r}.'
            message = message.format(type(self).__name__, argument)
            raise ValueError(message)
        if accidental is not None:
            accidental = abjad.Accidental(accidental)
            self._pitch_class = type(self._pitch_class)(
                self._pitch_class,
                accidental=accidental,
                )
        if arrow is not None:
            self._pitch_class = type(self._pitch_class)(
                self._pitch_class,
                arrow=arrow,
                )
        if octave is not None:
            octave = abjad.Octave(octave)
            self._octave = octave

    ### SPECIAL METHODS ###

    def __float__(self):
        """
        Coerce to float.

        Returns float.
        """
        return float(self.number)

    def __format__(self, format_specification=''):
        """
        Formats pitch.

        Set `format_specification` to `''`, `'lilypond'` or `'storage'`.

        Returns string.
        """
        if format_specification in ('', 'lilypond'):
            return self._get_lilypond_format()
        elif format_specification == 'storage':
            return StorageFormatManager(self).get_storage_format()
        return str(self)

    def __illustrate__(self):
        """
        Illustrates pitch.

        Returns LilyPond file.
        """
        import abjad
        pitch = abjad.NamedPitch(self)
        note = abjad.Note(pitch, 1)
        abjad.attach(abjad.Multiplier(1, 4), note)
        clef = abjad.Clef.from_selection([pitch])
        abjad.attach(clef, note)
        staff = abjad.Staff()
        staff.append(note)
        abjad.override(staff).time_signature.stencil = False
        lilypond_file = abjad.LilyPondFile.new(staff)
        return lilypond_file

    @abc.abstractmethod
    def __lt__(self, argument):
        """
        Is true when pitch is less than `argument`.

        Returns true or false.
        """
        raise NotImplementedError

    ### PRIVATE PROPERTIES ###

    @abc.abstractmethod
    def _get_lilypond_format(self):
        raise NotImplementedError

    @staticmethod
    def _to_nearest_octave(pitch_number, pitch_class_number):
        target_pc = pitch_number % 12
        down = (target_pc - pitch_class_number) % 12
        up = (pitch_class_number - target_pc) % 12
        if up < down:
            return pitch_number + up
        else:
            return pitch_number - down

    @staticmethod
    def _to_nearest_quarter_tone(number):
        number = round(float(number) * 4) / 4
        div, mod = divmod(number, 1)
        if mod == 0.75:
            div += 1
        elif mod == 0.5:
            div += 0.5
        return mathtools.integer_equivalent_number_to_integer(div)

    @staticmethod
    def _to_pitch_class_item_class(item_class):
        import abjad
        item_class = item_class or abjad.NumberedPitch
        if item_class in (abjad.NamedPitchClass, abjad.NumberedPitchClass):
            return item_class
        elif item_class is abjad.NamedPitch:
            return abjad.NamedPitchClass
        elif item_class is abjad.NumberedPitch:
            return abjad.NumberedPitchClass
        else:
            raise TypeError(item_class)

    @staticmethod
    def _to_pitch_item_class(item_class):
        import abjad
        item_class = item_class or abjad.NumberedPitch
        if item_class in (abjad.NamedPitch, abjad.NumberedPitch):
            return item_class
        elif item_class is abjad.NamedPitchClass:
            return abjad.NamedPitch
        elif item_class is abjad.NumberedPitchClass:
            return abjad.NumberedPitch
        else:
            raise TypeError(item_class)

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def arrow(self):
        """
        Gets arrow of pitch.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def hertz(self):
        """
        Gets frequency of pitch in Hertz.

        Returns float.
        """
        hertz = pow(2., (float(self.number) - 9.) / 12.) * 440.
        return hertz

    @abc.abstractproperty
    def name(self):
        """
        Gets name of pitch.

        Returns string.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def number(self):
        """
        Gets number of pitch.

        Returns number.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def octave(self):
        """
        Gets octave of pitch.

        Returns octave.
        """
        raise NotImplementedError

    @abc.abstractproperty
    def pitch_class(self):
        """
        Gets pitch-class of pitch.

        Returns pitch-class.
        """
        raise NotImplementedError

    ### PUBLIC METHODS ###

    @classmethod
    @abc.abstractmethod
    def from_hertz(class_, hertz):
        """
        Creates pitch from `hertz`.

        Returns new pitch.
        """
        midi = 9. + (12. * math.log(float(hertz) / 440., 2))
        pitch = class_(midi)
        return pitch

    @abc.abstractmethod
    def get_name(self, locale=None):
        """
        Gets name of pitch according to `locale`.

        Returns string.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def invert(self, axis=None):
        """
        Inverts pitch about `axis`.

        Interprets `axis` of none equal to middle C.

        Returns new pitch.
        """
        import abjad
        axis = axis or abjad.NamedPitch("c'")
        axis = type(self)(axis)
        interval = self - axis
        result = axis.transpose(interval)
        return result

    @abc.abstractmethod
    def multiply(self, n=1):
        """
        Multiplies pitch by `n`.

        Returns new pitch.
        """
        return type(self)(n * self.number)

    @abc.abstractmethod
    def transpose(self, n):
        """
        Transposes pitch by index `n`.

        Returns new pitch.
        """
        raise NotImplementedError



from abjad.system.AbjadValueObject import AbjadValueObject


class MeasureMaker(AbjadValueObject):
    r"""
    Measure-maker.

    ..  container:: example

        >>> maker = abjad.MeasureMaker()
        >>> measures = maker([(1, 8), (5, 16), (5, 16)])
        >>> staff = abjad.Staff(measures)
        >>> abjad.show(staff) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(staff)
            \new Staff
            {
                {   % measure
                    \time 1/8
                    s1 * 1/8
                }   % measure
                {   % measure
                    \time 5/16
                    s1 * 5/16
                }   % measure
                {   % measure
                    s1 * 5/16
                }   % measure
            }

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Makers'

    __slots__ = (
        '_implicit_scaling',
        )

    _publish_storage_format = True

    ### INITIALIZER ###

    def __init__(self, implicit_scaling=False):
        self._implicit_scaling = implicit_scaling

    ### SPECIAL METHODS ###

    def __call__(self, time_signatures):
        """
        Calls measure-maker on ``time_signatures``.

        Returns measures.
        """
        import abjad
        measures = []
        for time_signature in time_signatures:
            time_signature = abjad.TimeSignature(time_signature)
            measure = abjad.Measure(
                time_signature,
                implicit_scaling=self.implicit_scaling,
                )
            measures.append(measure)
        for i, measure in enumerate(measures):
            skip = abjad.Skip(1)
            # allow zero-update iteration
            time_signature = measure.time_signature
            duration = time_signature.duration
            if measure.implicit_scaling:
                implied_prolation = time_signature.implied_prolation
                multiplier = duration.__div__(implied_prolation)
            else:
                multiplier = abjad.Multiplier(duration)
            abjad.attach(multiplier, skip)
            measure[:] = [skip]
            # REMOVE: spanners attach only to leaves:
            #for spanner in measure._get_spanners():
            #    spanner._remove(measure)
        return abjad.select(measures)

    ### PUBLIC PROPERTIES ###

    @property
    def implicit_scaling(self):
        """
        Is true when measure scale implicitly without top-level tuplet.

        Returns true, false or none.
        """
        return self._implicit_scaling



import typing
from abjad.system.AbjadValueObject import AbjadValueObject
from abjad.system.LilyPondFormatBundle import LilyPondFormatBundle


class LilyPondComment(AbjadValueObject):
    r"""
    LilyPond comment.

    ..  container:: example

        Two-word comment:

        >>> note = abjad.Note("c'4")
        >>> comment = abjad.LilyPondComment('a comment')
        >>> abjad.attach(comment, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(note)
            % a comment
            c'4

    ..  container:: example

        Three-word comment:

        >>> note = abjad.Note("c'4")
        >>> comment = abjad.LilyPondComment('yet another comment')
        >>> abjad.attach(comment, note)
        >>> abjad.show(note) # doctest: +SKIP

        ..  docs::

            >>> abjad.f(note)
            % yet another comment
            c'4

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        '_string',
        '_format_slot',
        )

    _allowable_format_slots = (
        'after',
        'before',
        'closing',
        'opening',
        )

    _can_attach_to_containers = True

    _format_leaf_children = False

    ### INITIALIZER ###

    def __init__(
        self,
        string: str = None,
        format_slot: str = 'before',
        ) -> None:
        if isinstance(string, type(self)):
            argument = string
            string = argument.string
            format_slot = format_slot or argument.format_slot
        else:
            string = str(string)
        self._string = string
        assert format_slot in self._allowable_format_slots, repr(format_slot)
        self._format_slot = format_slot

    ### SPECIAL METHODS ###

    def __str__(self) -> str:
        """
        Gets string representation of LilyPond comment.

        ..  container:: example

            Two-word comment:

            >>> comment = abjad.LilyPondComment('a comment')
            >>> str(comment)
            '% a comment'

        ..  container:: example

            Three-word comment:

            >>> comment = abjad.LilyPondComment('yet another comment')
            >>> str(comment)
            '% yet another comment'

        """
        return rf'% {self.string}'

    ### PRIVATE METHODS ###

    def _get_lilypond_format(self):
        return str(self)

    def _get_lilypond_format_bundle(self, component=None):
        bundle = LilyPondFormatBundle()
        format_slot = bundle.get(self.format_slot)
        format_slot.comments.append(self._get_lilypond_format())
        return bundle

    ### PUBLIC PROPERTIES ###

    @property
    def string(self) -> str:
        """
        Gets string.

        ..  container:: example

            Two-word comment:

            >>> comment = abjad.LilyPondComment('a comment')
            >>> comment.string
            'a comment'

        ..  container:: example

            Three-word comment:

            >>> comment = abjad.LilyPondComment('yet another comment')
            >>> comment.string
            'yet another comment'

        """
        return self._string

    @property
    def format_slot(self) -> str:
        """
        Format slot of LilyPond comment.

        ..  container:: example

            Two-word comment:

            >>> comment = abjad.LilyPondComment('a comment')
            >>> comment.format_slot
            'before'

        ..  container:: example

            Three-word comment:

            >>> comment = abjad.LilyPondComment('yet another comment')
            >>> comment.format_slot
            'before'

        """
        return self._format_slot

    @property
    def tweaks(self) -> None:
        """
        Are not implemented on LilyPond comment.
        """
        pass

    ### PUBLIC METHODS ###

    @staticmethod
    def list_allowable_format_slots() -> typing.Tuple[str, ...]:
        """
        Lists allowable format slots.

        ..  container:: example

            >>> commands = abjad.LilyPondComment.list_allowable_format_slots()
            >>> for command in commands:
            ...     command
            ...
            'after'
            'before'
            'closing'
            'opening'

        """
        return LilyPondComment._allowable_format_slots



import pathlib
import re
import roman  # type: ignore
import six
import textwrap
import typing
import unicodedata
from abjad import enums
from .TypedList import TypedList


class String(str):
    """
    String.
    """

    ### CLASS VARIABLES ###

    hyphen_delimited_lowercase_regex_body = '(([a-z,0-9]+[-]+)*[a-z,0-9]+)?'
    hyphen_delimited_lowercase_regex = re.compile(
        '^{}$'.format(hyphen_delimited_lowercase_regex_body),
        re.VERBOSE,
        )

    hyphen_delimited_lowercase_file_name_regex_body = """
        {}
        (\.[a-z,0-9]+)?
        """.format(hyphen_delimited_lowercase_regex_body)

    hyphen_delimited_lowercase_file_name_regex = re.compile(
        '^{}$'.format(hyphen_delimited_lowercase_file_name_regex_body),
        re.VERBOSE,
        )

    lowercamelcase_regex = re.compile(
        '^([a-z,0-9]+([A-Z,0-9]+[a-z,0-9]*)*)?$',
        re.VERBOSE,
        )

    space_delimited_lowercase_regex = re.compile(
        '^(([a-z,0-9]+[ ]+)*[a-z,0-9]+)?$',
        re.VERBOSE,
        )

    underscore_delimited_lowercase_regex_body = '(([a-z,0-9]+[_]+)*[a-z,0-9]+)?'
    underscore_delimited_lowercase_regex = re.compile(
        '^{}$'.format(underscore_delimited_lowercase_regex_body),
        re.VERBOSE,
        )

    underscore_delimited_lowercase_file_name_regex_body = """
        {}
        (\.[a-z,0-9]+)?
        """.format(underscore_delimited_lowercase_regex_body)

    underscore_delimited_lowercase_file_name_regex = re.compile(
        '^{}$'.format(underscore_delimited_lowercase_file_name_regex_body),
        re.VERBOSE,
        )

    underscore_delimited_lowercase_file_name_with_extension_regex_body = """
        {}
        \.
        [a-z,0-9]+
        """.format(underscore_delimited_lowercase_regex_body)

    underscore_delimited_lowercase_file_name_with_extension_regex = re.compile(
        '^{}$'.format(underscore_delimited_lowercase_file_name_with_extension_regex_body),
        re.VERBOSE,
        )

    underscore_delimited_lowercase_package_regex_body = """
        ({}\.)*
        {}
        """.format(
            underscore_delimited_lowercase_regex_body,
            underscore_delimited_lowercase_regex_body,
            )

    underscore_delimited_lowercase_package_regex = re.compile(
        '^{}$'.format(underscore_delimited_lowercase_package_regex_body),
        re.VERBOSE,
        )

    uppercamelcase_regex = re.compile(
        '^([A-Z,0-9]+[a-z,0-9]*)*$',
        re.VERBOSE,
        )

    ### PRIVATE METHODS ###

    def _is_wrapper_directory_name(self):
        if self in ('.git', '.DS_Store'):
            return False
        return True

    ### PUBLIC METHODS ###

    @staticmethod
    def base_26(n: int) -> 'String':
        """
        Gets base-26 representation of nonnegative integer ``n``.

        :param n: integer.

        ..  container:: example

            >>> abjad.String.base_26(1)
            'A'
            >>> abjad.String.base_26(2)
            'B'
            >>> abjad.String.base_26(3)
            'C'

            >>> abjad.String.base_26(26)
            'Z'
            >>> abjad.String.base_26(27)
            'AA'
            >>> abjad.String.base_26(28)
            'AB'

            >>> abjad.String.base_26(52)
            'AZ'
            >>> abjad.String.base_26(53)
            'BA'
            >>> abjad.String.base_26(54)
            'BB'

            >>> abjad.String.base_26(78)
            'BZ'
            >>> abjad.String.base_26(79)
            'CA'
            >>> abjad.String.base_26(80)
            'CB'

        """
        assert 0 < n, repr(n)
        if 1 <= n <= 26:
            result = chr(ord('A') + n - 1)
        elif 26 < n < 676:
            left = int(n / 26)
            right = n - (26 * left)
            if right == 0:
                left -= 1
                right = 26
            left_ = chr(ord('A') + left - 1)
            right_ = chr(ord('A') + right - 1)
            result = left_ + right_
        else:
            raise NotImplementedError(n)
        return String(result)

    def capitalize_start(self) -> 'String':
        """
        Capitalizes start of string.

        ..  container:: example

            >>> abjad.String('violin I').capitalize_start()
            'Violin I'

        Capitalizes only ``string[0]``; leaves noninitial characters unchanged.

        ..  container:: example

            Built-in ``string.capitalize()`` forces noninitial characters to
            lowercase:

            >>> 'violin I'.capitalize()
            'Violin i'

        """
        if not self:
            return type(self)('')
        return type(self)(self[0].upper() + self[1:])

    def delimit_words(self, separate_caps=False) -> typing.List['String']:
        """
        Delimits words in string.

        ..  container:: example

            >>> string = abjad.String('scale degrees 4 and 5.')
            >>> string.delimit_words()
            ['scale', 'degrees', '4', 'and', '5']

            >>> string = abjad.String('scale degrees 4and5.')
            >>> string.delimit_words()
            ['scale', 'degrees', '4', 'and', '5']

            >>> string = abjad.String('scaleDegrees4and5.')
            >>> string.delimit_words()
            ['scale', 'Degrees', '4', 'and', '5']

            >>> string = abjad.String('ScaleDegrees4and 5.')
            >>> string.delimit_words()
            ['Scale', 'Degrees', '4', 'and', '5']

            >>> string = abjad.String('scale-degrees-4-and-5.')
            >>> string.delimit_words()
            ['scale', 'degrees', '4', 'and', '5']

            >>> string = abjad.String('SCALE_DEGREES_4_AND_5.')
            >>> string.delimit_words()
            ['SCALE', 'DEGREES', '4', 'AND', '5']

        ..  container:: example

            >>> string = abjad.String('one < two')
            >>> string.delimit_words()
            ['one', '<', 'two']

            >>> string = abjad.String('one! two!')
            >>> string.delimit_words()
            ['one', '!', 'two', '!']

        ..  container:: example

            Separates capital letters when keyword is true:

            >>> string = abjad.String('MRM')
            >>> string.delimit_words()
            ['MRM']
            >>> string.delimit_words(separate_caps=True)
            ['M', 'R', 'M']

            >>> string = abjad.String('MRhM')
            >>> string.delimit_words()
            ['MRh', 'M']
            >>> string.delimit_words(separate_caps=True)
            ['M', 'Rh', 'M']

        """
        wordlike_characters = ('<', '>', '!')
        words = []
        current_word = ''
        for character in self:
            if (not character.isalpha() and
                not character.isdigit() and
                character not in wordlike_characters):
                if current_word:
                    words.append(current_word)
                    current_word = ''
            elif not current_word:
                current_word = current_word + character
            elif character.isupper():
                if current_word[-1].isupper() and not separate_caps:
                    current_word = current_word + character
                else:
                    words.append(current_word)
                    current_word = character
            elif character.islower():
                if current_word[-1].isalpha():
                    current_word = current_word + character
                else:
                    words.append(current_word)
                    current_word = character
            elif character.isdigit():
                if current_word[-1].isdigit():
                    current_word = current_word + character
                else:
                    words.append(current_word)
                    current_word = character
            elif character in wordlike_characters:
                if current_word[-1] in wordlike_characters:
                    current_word = current_word + character
                else:
                    words.append(current_word)
                    current_word = character
        if current_word:
            words.append(current_word)
        return [type(self)(_) for _ in words]

    def from_roman(self) -> int:
        """
        Changes string from Roman numeral to digits.

        ..  container:: example

            >>> abjad.String('IX').from_roman()
            9
            >>> abjad.String('ix').from_roman()
            9

        ..  container:: example

            Raises Roman numeral error when string is not Roman numeral:

            >>> abjad.String('Allegro').from_roman()
            Traceback (most recent call last):
                ...
            roman.InvalidRomanNumeralError: Invalid Roman numeral: Allegro

        """
        if self.is_roman():
            return roman.fromRoman(self.upper())
        number = roman.fromRoman(self)
        return number

    def is_build_directory_name(self) -> bool:
        """
        Is true when string is build directory name.

        ..  container:: example

            >>> abjad.String('letter-score').is_build_directory_name()
            True
            >>> abjad.String('letter_score').is_build_directory_name()
            True

            >>> abjad.String('Letter_score').is_build_directory_name()
            False
            >>> abjad.String('-letter-score').is_build_directory_name()
            False
            >>> abjad.String('_letter-score').is_build_directory_name()
            False
            >>> abjad.String('letter score').is_build_directory_name()
            False

        """
        if not self == self.lower():
            return False
        if ' ' in self:
            return False
        if not self[0].isalpha():
            return False
        return True

    def is_classfile_name(self) -> bool:
        """
        Is true when string is classfile name.

        ..  container:: example

            >>> abjad.String('BeamSpecifier.py').is_classfile_name()
            True

            >>> abjad.String('BeamSpecifier').is_classfile_name()
            False
            >>> abjad.String('beamSpecifier.py').is_classfile_name()
            False
            >>> abjad.String('Beam_Specifier.py').is_classfile_name()
            False

        """
        path = pathlib.Path(self)
        if not type(self)(path.stem).is_upper_camel_case():
            return False
        if not path.suffix == '.py':
            return False
        return True

    def is_dash_case(self) -> bool:
        """
        Is true when string is hyphen-delimited lowercase.

        ..  container:: example

            >>> abjad.String('foo-bar').is_dash_case()
            True

            >>> abjad.String('foo bar').is_dash_case()
            False

        """
        return bool(String.hyphen_delimited_lowercase_regex.match(self))

    def is_dash_case_file_name(self) -> bool:
        """
        Is true when string is hyphen-delimited lowercase file name with
        extension.

        ..  container:: example

            >>> abjad.String('foo-bar').is_dash_case_file_name()
            True

            >>> abjad.String('foo.bar.blah').is_dash_case_file_name()
            False

        """
        if self == '':
            return True
        return bool(
            String.hyphen_delimited_lowercase_file_name_regex.match(self)
            )

    def is_introduction_segment_name(self) -> bool:
        """
        Is true when string is introductory segment name.

        ..  container:: example

            >>> abjad.String('_').is_introduction_segment_name()
            True
            >>> abjad.String('_1').is_introduction_segment_name()
            True
            >>> abjad.String('_2').is_introduction_segment_name()
            True
            >>> abjad.String('_99').is_introduction_segment_name()
            True

            >>> abjad.String('__').is_introduction_segment_name()
            False
            >>> abjad.String('__1').is_introduction_segment_name()
            False

            >>> abjad.String('A').is_introduction_segment_name()
            False
            >>> abjad.String('1').is_introduction_segment_name()
            False

        """
        if not self:
            return False
        if self[0] != '_':
            return False
        if bool(self[1:]) and not self[1:].isdigit():
            return False
        return True

    def is_lilypond_identifier(self) -> bool:
        """
        Is true when string starts with a letter and contains only letters
        and underscores thereafter.

        ..  container:: example

            >>> abjad.String('ViolinOne').is_lilypond_identifier()
            True
            >>> abjad.String('Violin_One').is_lilypond_identifier()
            True

            >>> abjad.String('Violin One').is_lilypond_identifier()
            False

        ..  container:: example

            >>> abjad.String('ViolinI').is_lilypond_identifier()
            True
            >>> abjad.String('Violin_I').is_lilypond_identifier()
            True

            >>> abjad.String('Violin I').is_lilypond_identifier()
            False

        ..  container:: example

            >>> abjad.String('Violin1').is_lilypond_identifier()
            False
            >>> abjad.String('Violin_1').is_lilypond_identifier()
            False
            >>> abjad.String('Violin 1').is_lilypond_identifier()
            False

        """
        if self and self[0] == '_':
            return False
        for character in self:
            if not (character.isalpha() or character == '_'):
                return False
        return True

    def is_lower_camel_case(self) -> bool:
        """
        Is true when string and is lower camel case.

        ..  container:: example

            >>> abjad.String('fooBar').is_lower_camel_case()
            True

            >>> abjad.String('FooBar').is_lower_camel_case()
            False

        """
        return bool(String.lowercamelcase_regex.match(self))

    def is_lowercase_file_name(self) -> bool:
        """
        Is true when string is lowercase file name.

        ..  container:: example

            Is true when string is dash case:

            >>> abjad.String('back-cover').is_lowercase_file_name()
            True
            >>> abjad.String('back-cover.tex').is_lowercase_file_name()
            True

            Is true when string is underscore-delimited:

            >>> abjad.String('compile_source').is_lowercase_file_name()
            True
            >>> abjad.String('compile_source.py').is_lowercase_file_name()
            True

        ..  container:: example

            >>> abjad.String('back cover').is_lowercase_file_name()
            False
            >>> abjad.String('Back-Cover').is_lowercase_file_name()
            False

        """
        if not self == self.lower():
            return False
        path = pathlib.Path(self)
        if not (type(self)(path.stem).is_snake_case() or
            type(self)(path.stem).is_dash_case()):
            return False
        return True

    def is_module_file_name(self) -> bool:
        """
        Is true when string is module file name.

        ..  container:: example

            >>> abjad.String('compile_source.py').is_module_file_name()
            True

            >>> abjad.String('compile_source').is_module_file_name()
            False
            >>> abjad.String('compile-source').is_module_file_name()
            False
            >>> abjad.String('compile-source.py').is_module_file_name()
            False

            >>> abjad.String('Compile_Source.py').is_module_file_name()
            False

        """
        path = pathlib.Path(self)
        if not path.name == path.name.lower():
            return False
        if not type(self)(path.stem).is_snake_case():
            return False
        if not path.suffix == '.py':
            return False
        return True

    def is_package_name(self) -> bool:
        """
        Is true when string is package name.

        ..  container:: example

            >>> abjad.String('pitches').is_package_name()
            True
            >>> abjad.String('pitch_classes').is_package_name()
            True

            >>> abjad.String('Pitches').is_package_name()
            False
            >>> abjad.String('pitch-classes').is_package_name()
            False

        """
        if not self == self.lower():
            return False
        if not self.is_snake_case():
            return False
        return True

    def is_public_python_file_name(self) -> bool:
        """
        Is true when string is public Python file name.

        ..  container:: example

            >>> abjad.String('pitches.py').is_public_python_file_name()
            True
            >>> abjad.String('pitch_classes.py').is_public_python_file_name()
            True
            >>> abjad.String('Pitches.py').is_public_python_file_name()
            True
            >>> abjad.String('PitchClasses.py').is_public_python_file_name()
            True

            >>> abjad.String('.pitches.py').is_public_python_file_name()
            False
            >>> abjad.String('_pitches.py').is_public_python_file_name()
            False
            >>> abjad.String('pitch-classes.py').is_public_python_file_name()
            False

        """
        path = pathlib.Path(self)
        if path.stem and not path.stem[0].isalpha():
            return False
        if not path.suffix == '.py':
            return False
        for character in path.stem[1:]:
            if not character.isalpha() and character != '_':
                return False
        return True

    def is_rehearsal_mark(self, forbid_i=False) -> bool:
        """
        Is true when string is one- or two-character rehearsal mark.

        ..  container:: example

            >>> abjad.String('A').is_rehearsal_mark()
            True
            >>> abjad.String('B').is_rehearsal_mark()
            True
            >>> abjad.String('Z').is_rehearsal_mark()
            True
            >>> abjad.String('AA').is_rehearsal_mark()
            True
            >>> abjad.String('AB').is_rehearsal_mark()
            True
            >>> abjad.String('AZ').is_rehearsal_mark()
            True
            >>> abjad.String('ZZ').is_rehearsal_mark()
            True

            >>> abjad.String('A1').is_rehearsal_mark()
            False
            >>> abjad.String('AAA').is_rehearsal_mark()
            False

        ..  container:: example

            >>> abjad.String('I').is_rehearsal_mark()
            True
            >>> abjad.String('J').is_rehearsal_mark()
            True

        ..  container:: example

            >>> abjad.String('I').is_rehearsal_mark(forbid_i=True)
            False

            >>> abjad.String('J').is_rehearsal_mark(forbid_i=True)
            True

        """
        if len(self) not in (1, 2):
            return False
        if not self.isalpha():
            return False
        if not self.isupper():
            return False
        if forbid_i is True and 'I' in self:
            return False
        return True

    def is_roman(self) -> bool:
        """
        Is true when string is Roman numeral.

        ..  container:: example

            >>> abjad.String('I').is_roman()
            True
            >>> abjad.String('II').is_roman()
            True
            >>> abjad.String('X').is_roman()
            True
            >>> abjad.String('XI').is_roman()
            True
            >>> abjad.String('C').is_roman()
            True
            >>> abjad.String('CI').is_roman()
            True

            >>> abjad.String('i').is_roman()
            True

            >>> abjad.String('F').is_roman()
            False

        """
        try:
            roman.fromRoman(self.upper())
            return True
        except roman.InvalidRomanNumeralError:
            return False

    def is_segment_name(self) -> bool:
        """
        Is true when string is segment name or package name.

        ..  container:: example

            >>> abjad.String('_').is_segment_name()
            True
            >>> abjad.String('_1').is_segment_name()
            True
            >>> abjad.String('_99').is_segment_name()
            True

        ..  container:: example

            >>> abjad.String('A').is_segment_name()
            True
            >>> abjad.String('A1').is_segment_name()
            True
            >>> abjad.String('A99').is_segment_name()
            True

        ..  container:: example

            >>> abjad.String('AB').is_segment_name()
            True
            >>> abjad.String('AB1').is_segment_name()
            True
            >>> abjad.String('AB99').is_segment_name()
            True

        ..  container:: example

            >>> abjad.String('segment_01').is_segment_name()
            True

        ..  container:: example

            >>> abjad.String('__1').is_segment_name()
            False
            >>> abjad.String('A_1').is_segment_name()
            False
            >>> abjad.String('AB_1').is_segment_name()
            False

        """
        if self.is_package_name():
            return True
        if self.is_introduction_segment_name():
            return True
        if self.is_rehearsal_mark():
            return True
        if String(self[0]).is_rehearsal_mark() and self[1:].isdigit():
            return True
        if String(self[:2]).is_rehearsal_mark() and self[2:].isdigit():
            return True
        return False

    def is_shout_case(self) -> bool:
        """
        Is true when string and is shoutcase.

        ..  container:: example

            >>> abjad.String('FOO_BAR').is_shout_case()
            True

            >>> abjad.String('FooBar').is_shout_case()
            False

        """
        return self == self.to_shout_case()

    def is_snake_case(self) -> bool:
        """
        Is true when string is underscore-delimited lowercase.

        ..  container:: example

            >>> abjad.String('foo_bar').is_snake_case()
            True

            >>> abjad.String('foo bar').is_snake_case()
            False

        """
        return bool(String.underscore_delimited_lowercase_regex.match(self))

    def is_snake_case_file_name(self) -> bool:
        """
        Is true when string is underscore-delimited lowercase file name.

        ..  container:: example

            >>> abjad.String('foo_bar').is_snake_case_file_name()
            True

            >>> abjad.String('foo.bar.blah').is_snake_case_file_name()
            False

        """
        if self == '':
            return True
        return bool(
            String.underscore_delimited_lowercase_file_name_regex.match(self)
            )

    def is_snake_case_file_name_with_extension(self) -> bool:
        """
        Is true when string is underscore-delimited lowercase file name
        with extension.

        ..  container:: example

            >>> string = abjad.String('foo_bar.blah')
            >>> string.is_snake_case_file_name_with_extension()
            True

            >>> string = abjad.String('foo.bar.blah')
            >>> string.is_snake_case_file_name_with_extension()
            False

        """
        if self == '':
            return True
        return bool(
            String.underscore_delimited_lowercase_file_name_with_extension_regex.match(
                self))

    def is_snake_case_package_name(self) -> bool:
        """
        Is true when string is underscore-delimited lowercase package name.

        ..  container:: example

            >>> string = abjad.String('foo.bar.blah_package')
            >>> string.is_snake_case_package_name()
            True

            >>> string = abjad.String('foo.bar.BlahPackage')
            >>> string.is_snake_case_package_name()
            False

        """
        return bool(
            String.underscore_delimited_lowercase_package_regex.match(self)
            )

    def is_space_delimited_lowercase(self) -> bool:
        """
        Is true when string is space-delimited lowercase.

        ..  container:: example

            >>> abjad.String('foo bar').is_space_delimited_lowercase()
            True

            >>> abjad.String('foo_bar').is_space_delimited_lowercase()
            False

        """
        return bool(String.space_delimited_lowercase_regex.match(self))

    @staticmethod
    def is_string(argument) -> bool:
        """
        Is true when ``argument`` is a string.

        ..  container:: example

            >>> abjad.String.is_string('Allegro')
            True
            >>> abjad.String.is_string('')
            True

            >>> abjad.String.is_string(99)
            False

        """
        return isinstance(argument, six.string_types)

    def is_stylesheet_name(self) -> bool:
        """
        Is true when string is stylesheet name.

        ..  container:: example

            >>> abjad.String('articulations.ily').is_stylesheet_name()
            True

            >>> abjad.String('Articulations.ily').is_stylesheet_name()
            False
            >>> abjad.String('articulations.ly').is_stylesheet_name()
            False
            >>> abjad.String('articulations').is_stylesheet_name()
            False
            >>> abjad.String('_articulations.ily').is_stylesheet_name()
            False

        """
        path = pathlib.Path(self)
        if not path.name == path.name.lower():
            return False
        if not type(self)(path.stem).is_dash_case():
            return False
        if not path.suffix == '.ily':
            return False
        return True

    def is_tools_file_name(self) -> bool:
        """
        Is true when string is tools file name.

        ..  container:: example

            >>> abjad.String('beam_specifier.py').is_tools_file_name()
            True
            >>> abjad.String('BeamSpecifier.py').is_tools_file_name()
            True

            >>> abjad.String('BeamSpecifier').is_tools_file_name()
            False
            >>> abjad.String('beamSpecifier.py').is_tools_file_name()
            False
            >>> abjad.String('Beam_Specifier.py').is_tools_file_name()
            False

        """
        if self.is_classfile_name():
            return True
        if self.is_module_file_name():
            return True
        return False

    def is_upper_camel_case(self) -> bool:
        """
        Is true when string upper camel case.

        ..  container:: example

            >>> abjad.String('FooBar').is_upper_camel_case()
            True

            >>> abjad.String('fooBar').is_upper_camel_case()
            False

        """
        return bool(String.uppercamelcase_regex.match(self))

    @staticmethod
    def match_strings(strings, pattern) -> typing.List[int]:
        """
        Matches ``pattern`` against ``strings``.

        ..  container:: example

            >>> strings = [
            ...     'AcciaccaturaSpecifier.py',
            ...     'AnchorCommand.py',
            ...     'ArpeggiationSpacingSpecifier.py',
            ...     'AttachCommand.py',
            ...     'ChordalSpacingSpecifier.py',
            ...     ]

            >>> abjad.String.match_strings(strings, 'A')
            []

            >>> for i in abjad.String.match_strings(strings, 'At'):
            ...     strings[i]
            'AttachCommand.py'

            >>> for i in abjad.String.match_strings(strings, 'AtC'):
            ...     strings[i]
            'AttachCommand.py'

            >>> for i in abjad.String.match_strings(strings, 'ASS'):
            ...     strings[i]
            'ArpeggiationSpacingSpecifier.py'

            >>> for i in abjad.String.match_strings(strings, 'AC'):
            ...     strings[i]
            'AnchorCommand.py'
            'AttachCommand.py'

            >>> for i in abjad.String.match_strings(strings, '.py'):
            ...     strings[i]
            'AcciaccaturaSpecifier.py'
            'AnchorCommand.py'
            'ArpeggiationSpacingSpecifier.py'
            'AttachCommand.py'
            'ChordalSpacingSpecifier.py'

            >>> abjad.String.match_strings(strings, '@AC')
            []

        ..  container:: example

            REGRESSION:

            >>> abjad.String.match_strings(strings, '||')
            []

            >>> strings = ['_allegro', '-allegro', '.allegro', 'allegro']
            >>> for i in abjad.String.match_strings(strings, '_al'):
            ...     strings[i]
            '_allegro'

            >>> for i in abjad.String.match_strings(strings, '-al'):
            ...     strings[i]
            '-allegro'

            >>> for i in abjad.String.match_strings(strings, '.al'):
            ...     strings[i]
            '.allegro'

            >>> for i in abjad.String.match_strings(strings, 'egro'):
            ...     strings[i]
            '_allegro'
            '-allegro'
            '.allegro'
            'allegro'

        """
        if not pattern:
            return []
        if not pattern[0].isalpha() and not pattern[0] in list('_-.'):
            return []
        pattern = String(pattern)
        indices = []
        for i, string in enumerate(strings):
            if string == pattern:
                indices.append(i)
        strings = [String(_) for _ in strings]
        if 3 <= len(pattern):
            for i, string in enumerate(strings):
                if string.startswith(pattern):
                    if i not in indices:
                        indices.append(i)
            for i, string in enumerate(strings):
                string = string.strip_diacritics().lower()
                if string.startswith(pattern.lower()):
                    if i not in indices:
                        indices.append(i)
        if len(pattern) <= 1:
            return indices
        if not pattern.islower() or any(_.isdigit() for _ in pattern):
            pattern_words = pattern.delimit_words(separate_caps=True)
            if pattern_words:
                for i, string in enumerate(strings):
                    if (string.startswith(pattern_words[0]) and
                        string.match_word_starts(pattern_words)):
                        if i not in indices:
                            indices.append(i)
                for i, string in enumerate(strings):
                    if string.match_word_starts(pattern_words):
                        if i not in indices:
                            indices.append(i)
        if pattern.islower():
            pattern_characters = list(pattern)
            if pattern_characters:
                for i, string in enumerate(strings):
                    if (string.startswith(pattern_characters[0]) and
                        string.match_word_starts(pattern_characters)):
                        if i not in indices:
                            indices.append(i)
                for i, string in enumerate(strings):
                    if string.match_word_starts(pattern_characters):
                        if i not in indices:
                            indices.append(i)
        if len(pattern) < 3:
            return indices
        for i, string in enumerate(strings):
            if pattern in string.strip_diacritics().lower():
                if i not in indices:
                    indices.append(i)
        return indices

    def match_word_starts(self, words) -> bool:
        """
        Matches word starts.

        ..  container:: example

            >>> string = abjad.String('StringQuartetScoreTemplate')

            >>> string.match_word_starts(['Str', 'Quar', 'Temp'])
            True

            >>> string.match_word_starts(['S', 'Q', 'S', 'T'])
            True

            >>> string.match_word_starts(['Quartet'])
            True

            >>> string.match_word_starts(['Q'])
            True

        ..  container:: example

            >>> string = abjad.String('StringQuartetScoreTemplate')

            >>> string.match_word_starts(['Quar', 'Str'])
            False

            >>> string.match_word_starts(['uartet'])
            False

        ..  container:: example

            REGRESSION:

            >>> string = abjad.String('IOManager')
            >>> string.match_word_starts(['I', 'O'])
            True

            >>> string = abjad.String('StringQuartetScoreTemplate')
            >>> string.match_word_starts(['S', 'Q', 'S', 'T', 'T'])
            False

            >>> string = abjad.String('AcciaccaturaSpecifier.py')
            >>> string.match_word_starts(['A', 'S', 'S'])
            False

        """
        my_words = self.delimit_words(separate_caps=True)
        indices: typing.List[int] = []
        last_index = 0
        for word in words:
            for i, my_word in enumerate(my_words):
                if i < last_index:
                    continue
                if my_word.startswith(word) and i not in indices:
                    indices.append(i)
                    last_index = i
                    break
            else:
                return False
        return indices == sorted(indices)

    @staticmethod
    def normalize(argument, indent=None) -> 'String':
        """
        Normalizes string.

        ..  container:: example

            >>> string = r'''
            ...     foo
            ...         bar
            ... '''
            >>> print(string)
            <BLANKLINE>
                foo
                    bar
            <BLANKLINE>

            >>> print(abjad.String.normalize(string))
            foo
                bar

            >>> print(abjad.String.normalize(string, indent=4))
                foo
                    bar

            >>> print(abjad.String.normalize(string, indent='* '))
            * foo
            *     bar

        """
        string = argument.replace('\t', '    ')
        lines = string.split('\n')
        while lines and (not lines[0] or lines[0].isspace()):
            lines.pop(0)
        while lines and (not lines[-1] or lines[-1].isspace()):
            lines.pop()
        for i, line in enumerate(lines):
            lines[i] = line.rstrip()
        string = '\n'.join(lines)
        string = textwrap.dedent(string)
        if indent:
            if not isinstance(indent, six.string_types):
                indent = ' ' * abs(int(indent))
            lines = string.split('\n')
            for i, line in enumerate(lines):
                if line:
                    lines[i] = f'{indent}{line}'
            string = '\n'.join(lines)
        return String(string)

    def pluralize(self, count=None) -> 'String':
        """
        Pluralizes English string.

        ..  container:: example

            Changes terminal ``-y`` to ``-ies``:

            >>> abjad.String('catenary').pluralize()
            'catenaries'

            Adds ``-es`` to terminal ``-s``, ``-sh``, ``-x`` and ``-z``:

            >>> abjad.String('brush').pluralize()
            'brushes'

            Adds ``-s`` to all other strings:

            >>> abjad.String('shape').pluralize()
            'shapes'

        ..  container:: example

            Does not pluralize when ``count`` is 1:

            >>> abjad.String('shape').pluralize(count=1)
            'shape'

        """
        if count == 1:
            return self
        elif self.endswith('y'):
            result = self[:-1] + 'ies'
        elif self.endswith(('s', 'sh', 'x', 'z')):
            result = self + 'es'
        else:
            result = self + 's'
        return type(self)(result)

    def remove_zfill(self) -> 'String':
        """
        Removes zfill from numbers in string.

        ..  container:: example

            >>> abjad.String('Horn1').remove_zfill()
            'Horn1'

            >>> abjad.String('Horn01').remove_zfill()
            'Horn1'

            >>> abjad.String('Horn001').remove_zfill()
            'Horn1'

        """
        words = []
        for word in self.delimit_words():
            try:
                number = int(word)
                words.append(str(number))
            except ValueError:
                words.append(word)
        result = ''.join(words)
        return type(self)(result)

    def segment_letter(self):
        """
        Gets segment letter.

        ..  container:: example

            >>> abjad.String('_').segment_letter()
            '_'
            >>> abjad.String('_1').segment_letter()
            '_'
            >>> abjad.String('_12').segment_letter()
            '_'

            >>> abjad.String('A').segment_letter()
            'A'
            >>> abjad.String('A1').segment_letter()
            'A'
            >>> abjad.String('A12').segment_letter()
            'A'

            >>> abjad.String('AB').segment_letter()
            'AB'
            >>> abjad.String('AB1').segment_letter()
            'AB'
            >>> abjad.String('AB12').segment_letter()
            'AB'

        """
        if not self.is_segment_name():
            raise ValueError(f'must be segment name (not {self!r}).')
        if len(self) == 1:
            return self
        elif len(self) == 2:
            if self[1].isdigit():
                return self[:1]
            else:
                return self
        elif len(self) == 3:
            if self[1].isdigit():
                return self[:1]
            else:
                return self[:2]
        else:
            return self[:2]

    def segment_rank(self):
        """
        Gets segment index.

        ..  container:: example

            >>> abjad.String('_').segment_rank()
            0
            >>> abjad.String('_1').segment_rank()
            1
            >>> abjad.String('_12').segment_rank()
            12

            >>> abjad.String('A').segment_rank()
            0
            >>> abjad.String('A1').segment_rank()
            1
            >>> abjad.String('A12').segment_rank()
            12

            >>> abjad.String('AB').segment_rank()
            0
            >>> abjad.String('AB1').segment_rank()
            1
            >>> abjad.String('AB12').segment_rank()
            12

        """
        if not self.is_segment_name():
            raise ValueError(f'must be segment name (not {self!r}).')
        prefix = 'segment_'
        if self.startswith(prefix):
            index = int(self[len(prefix):])
            return index
        letter = self.segment_letter()
        if letter == self:
            return 0
        index = int(self[len(letter):])
        return index

    @staticmethod
    def sort_roman(strings) -> typing.List['String']:
        """
        Sorts strings containing Roman numerals.
        
        ..  container:: example

            >>> strings = ['TromboneII', 'TromboneIII', 'TromboneI']
            >>> abjad.String.sort_roman(strings)
            ['TromboneI', 'TromboneII', 'TromboneIII']

            >>> strings = ['ViolinXI', 'ViolinX', 'ViolinIX']
            >>> abjad.String.sort_roman(strings)
            ['ViolinIX', 'ViolinX', 'ViolinXI']

        """
        lists = []
        for string in strings:
            list_ = []
            for word in String(string).delimit_words():
                try:
                    number = roman.fromRoman(word)
                    list_.append(number)
                except roman.InvalidRomanNumeralError:
                    list_.append(word)
            lists.append(list_)
        pairs = list(zip(strings, lists))
        pairs.sort(key=lambda pair: pair[1])
        strings_ = [pair[0] for pair in pairs]
        return strings_

    @staticmethod
    def sort_segment_names(strings) -> typing.List['String']:
        """
        Sorts segment name ``strings``.
        
        ..  container:: example

            >>> strings = ['AA', 'Z', '_11', '_9']
            >>> abjad.String.sort_segment_names(strings)
            ['_9', '_11', 'Z', 'AA']

        """
        names = []
        for string in strings:
            name = String(string)
            if not name.is_segment_name():
                raise ValueError(f'must be segment name (not {string!r}).')
            names.append(name)
        def _compare(name_1, name_2):
            letter_1 = name_1.segment_letter()
            letter_2 = name_2.segment_letter()
            rank_1 = name_1.segment_rank()
            rank_2 = name_2.segment_rank()
            if letter_1 == letter_2:
                if rank_1 < rank_2:
                    return -1
                if rank_1 == rank_2:
                    return 0
                if rank_1 > rank_2:
                    return 1
            if letter_1 == '_':
                return -1
            if letter_2 == '_':
                return 1
            if len(letter_1) == len(letter_2):
                if letter_1 < letter_2:
                    return -1
                if letter_2 < letter_1:
                    return 1
            if len(letter_1) < len(letter_2):
                return -1
            assert len(letter_2) < len(letter_1)
            return 1
        names_ = TypedList(names)
        names_.sort(cmp=_compare)
        return list(names_)

    def strip_diacritics(self) -> 'String':
        """
        Strips diacritics from string.

        ..  container:: example

            >>> string = abjad.String('Dvořák')
            >>> print(string)
            Dvořák

            >>> string.strip_diacritics()
            'Dvorak'

        """
        normalized_unicode_string = unicodedata.normalize('NFKD', self)
        ascii_string = normalized_unicode_string.encode('ascii', 'ignore')
        return type(self)(ascii_string.decode('utf-8'))

    def strip_roman(self) -> 'String':
        """
        Strips roman numerals from right of string.

        ..  container:: example

            >>> abjad.String('Trombone').strip_roman()
            'Trombone'

            >>> abjad.String('TromboneI').strip_roman()
            'Trombone'

            >>> abjad.String('TromboneII').strip_roman()
            'Trombone'

            >>> abjad.String('TromboneIII').strip_roman()
            'Trombone'

            >>> abjad.String('TromboneIV').strip_roman()
            'Trombone'

        """
        words = self.delimit_words()
        try:
            roman.fromRoman(words[-1])
            words = words[:-1]
        except roman.InvalidRomanNumeralError:
            pass
        return String(''.join(words))

    def to_accent_free_snake_case(self) -> 'String':
        """
        Changes string to accent-free snake case.

        ..  container:: example

            >>> abjad.String('Déja vu').to_accent_free_snake_case()
            'deja_vu'

        """
        string = self.strip_diacritics()
        string_ = string.replace(' ', '_')
        string_ = string_.replace("'", '_')
        string_ = string_.lower()
        return type(self)(string_)

    @staticmethod
    def to_bidirectional_direction_string(argument) -> 'String':
        """
        Changes ``argument`` to bidirectional direction string.

        ..  container:: example:

            >>> abjad.String.to_bidirectional_direction_string('^')
            'Up'

            >>> abjad.String.to_bidirectional_direction_string('_')
            'Down'

            >>> abjad.String.to_bidirectional_direction_string(1)
            'Up'

            >>> abjad.String.to_bidirectional_direction_string(-1)
            'Down'

        """
        try:
            alignment = enums.VerticalAlignment.from_expr(argument)
        except Exception:
            raise ValueError(repr(argument))
        if alignment is enums.Center:
            raise ValueError(repr(argument))
        return String(alignment)

    @staticmethod
    def to_bidirectional_lilypond_symbol(argument) -> 'String':
        """
        Changes ``argument`` to bidirectional LilyPond symbol.

        ..  container:: example

            >>> abjad.String.to_bidirectional_lilypond_symbol(abjad.Up)
            '^'

            >>> abjad.String.to_bidirectional_lilypond_symbol(abjad.Down)
            '_'

            >>> abjad.String.to_bidirectional_lilypond_symbol(1)
            '^'

            >>> abjad.String.to_bidirectional_lilypond_symbol(-1)
            '_'

        """
        try:
            alignment = enums.VerticalAlignment.from_expr(argument)
        except Exception:
            raise ValueError(repr(argument))
        if alignment is enums.Center:
            raise ValueError(repr(argument))
        return String(format(alignment, 'lilypond'))

    def to_dash_case(self) -> 'String':
        """
        Changes string to dash case.

        ..  container:: example

            >>> abjad.String('scale degrees 4 and 5').to_dash_case()
            'scale-degrees-4-and-5'

            >>> abjad.String('scale_degrees_4_and_5').to_dash_case()
            'scale-degrees-4-and-5'

            >>> abjad.String('scale-degrees-4-and-5').to_dash_case()
            'scale-degrees-4-and-5'

            >>> abjad.String('ScaleDegrees4And5').to_dash_case()
            'scale-degrees-4-and-5'

        """
        words = self.delimit_words()
        words_ = [_.lower() for _ in words]
        string = '-'.join(words_)
        return type(self)(string)

    @staticmethod
    def to_indicator_stem(indicator) -> 'String':
        """
        Changes ``indicator`` to stem.

        ..  container:: example

            >>> abjad.String.to_indicator_stem(abjad.Clef('alto'))
            'CLEF'

            >>> abjad.String.to_indicator_stem(abjad.Clef('treble'))
            'CLEF'

            >>> abjad.String.to_indicator_stem(abjad.Dynamic('f'))
            'DYNAMIC'

            >>> abjad.String.to_indicator_stem(abjad.Cello())
            'INSTRUMENT'

            >>> abjad.String.to_indicator_stem(abjad.Violin())
            'INSTRUMENT'

            >>> metronome_mark = abjad.MetronomeMark((1, 4), 58)
            >>> abjad.String.to_indicator_stem(metronome_mark)
            'METRONOME_MARK'

        """
        parameter = getattr(indicator, 'parameter', None)
        if isinstance(parameter, str):
            stem = parameter.lstrip('abjad.')
        else:
            stem = type(indicator).__name__
        return String(stem).to_shout_case()

    def to_lower_camel_case(self) -> 'String':
        """
        Changes string to lower camel case.

        ..  container:: example

            >>> abjad.String('scale degrees 4 and 5').to_lower_camel_case()
            'scaleDegrees4And5'

            >>> abjad.String('scale_degrees_4_and_5').to_lower_camel_case()
            'scaleDegrees4And5'

            >>> abjad.String('scale-degrees-4-and-5').to_lower_camel_case()
            'scaleDegrees4And5'

            >>> abjad.String('ScaleDegrees4And5').to_lower_camel_case()
            'scaleDegrees4And5'

        """
        string = self.to_upper_camel_case()
        if string == '':
            pass
        else:
            string = type(self)(string[0].lower() + string[1:])
        return string

    def to_segment_lilypond_identifier(self) -> 'String':
        """
        Changes string to segment LilyPond identifier.

        ..  container:: example

            >>> abjad.String('_').to_segment_lilypond_identifier()
            'i'
            >>> abjad.String('_1').to_segment_lilypond_identifier()
            'i_a'
            >>> abjad.String('_2').to_segment_lilypond_identifier()
            'i_b'

            >>> abjad.String('A').to_segment_lilypond_identifier()
            'A'
            >>> abjad.String('A1').to_segment_lilypond_identifier()
            'A_a'
            >>> abjad.String('A2').to_segment_lilypond_identifier()
            'A_b'

            >>> abjad.String('B').to_segment_lilypond_identifier()
            'B'
            >>> abjad.String('B1').to_segment_lilypond_identifier()
            'B_a'
            >>> abjad.String('B2').to_segment_lilypond_identifier()
            'B_b'

            >>> abjad.String('AA').to_segment_lilypond_identifier()
            'AA'
            >>> abjad.String('AA1').to_segment_lilypond_identifier()
            'AA_a'
            >>> abjad.String('AA2').to_segment_lilypond_identifier()
            'AA_b'

        """
        name = self.replace('_', 'i')
        words = []
        for word in String(name).delimit_words():
            if word.isdigit():
                word_ = String.base_26(int(word)).lower()
                words.append(word_)
            else:
                words.append(word)
        identifier = '_'.join(words)
        return String(identifier)

    def to_shout_case(self) -> 'String':
        """
        Changes string to shout case.

        ..  container:: example

            >>> abjad.String('scale degrees 4 and 5').to_shout_case()
            'SCALE_DEGREES_4_AND_5'

            >>> abjad.String('scale_degrees_4_and_5').to_shout_case()
            'SCALE_DEGREES_4_AND_5'

            >>> abjad.String('scale-degrees-4-and-5').to_shout_case()
            'SCALE_DEGREES_4_AND_5'

            >>> abjad.String('ScaleDegrees4And5').to_shout_case()
            'SCALE_DEGREES_4_AND_5'

        """
        words = self.delimit_words()
        words_ = [_.upper() for _ in words]
        string = '_'.join(words_)
        return type(self)(string)

    def to_snake_case(self) -> 'String':
        """
        Changes string to snake case.

        ..  container:: example

            >>> abjad.String('scale degrees 4 and 5').to_snake_case()
            'scale_degrees_4_and_5'

            >>> abjad.String('scale_degrees_4_and_5').to_snake_case()
            'scale_degrees_4_and_5'

            >>> abjad.String('scale-degrees-4-and-5').to_snake_case()
            'scale_degrees_4_and_5'

            >>> abjad.String('ScaleDegrees4And5').to_snake_case()
            'scale_degrees_4_and_5'

        """
        words = self.delimit_words()
        words_ = [_.lower() for _ in words]
        string = '_'.join(words_)
        return type(self)(string)

    def to_space_delimited_lowercase(self) -> 'String':
        """
        Changes string to space-delimited lowercase.

        ..  container:: example

            >>> abjad.String('LogicalTie').to_space_delimited_lowercase()
            'logical tie'

            >>> abjad.String('logical_tie').to_space_delimited_lowercase()
            'logical tie'

            >>> abjad.String('logical tie').to_space_delimited_lowercase()
            'logical tie'

        ..  container:: example

            >>> abjad.String('').to_space_delimited_lowercase()
            ''

        """
        if not self:
            return self
        elif self[0].isupper():
            words = []
            current_word = self[0].lower()
            for letter in self[1:]:
                if letter.isupper():
                    words.append(current_word)
                    current_word = letter.lower()
                else:
                    current_word = current_word + letter
            words.append(current_word)
            string = ' '.join(words)
        else:
            string = self.replace('_', ' ')
        return type(self)(string)

    @staticmethod
    def to_tridirectional_direction_string(argument) -> typing.Optional[
        'String']:
        """
        Changes ``argument`` to tridirectional direction string.

        ..  container:: example

            >>> abjad.String.to_tridirectional_direction_string('^')
            'Up'

            >>> abjad.String.to_tridirectional_direction_string('-')
            'Center'

            >>> abjad.String.to_tridirectional_direction_string('_')
            'Down'

            >>> abjad.String.to_tridirectional_direction_string(1)
            'Up'

            >>> abjad.String.to_tridirectional_direction_string(0)
            'Center'

            >>> abjad.String.to_tridirectional_direction_string(-1)
            'Down'

        ..  container:: example

            Returns none when ``argument`` is none:

            >>> abjad.String.to_tridirectional_direction_string(None) is None
            True

        """
        if argument is None:
            return None
        try:
            alignment = enums.VerticalAlignment.from_expr(argument)
        except Exception:
            raise ValueError(repr(argument))
        return String(alignment)

    @staticmethod
    def to_tridirectional_lilypond_symbol(
        argument: typing.Any,
        ) -> typing.Optional['String']:
        """
        Changes ``argument`` to tridirectional LilyPond symbol.

        ..  container:: example

            >>> abjad.String.to_tridirectional_lilypond_symbol(abjad.Up)
            '^'

            >>> abjad.String.to_tridirectional_lilypond_symbol('center')
            '-'

            >>> abjad.String.to_tridirectional_lilypond_symbol(abjad.Down)
            '_'

            >>> abjad.String.to_tridirectional_lilypond_symbol(1)
            '^'

            >>> abjad.String.to_tridirectional_lilypond_symbol(0)
            '-'

            >>> abjad.String.to_tridirectional_lilypond_symbol(-1)
            '_'

            >>> abjad.String.to_tridirectional_lilypond_symbol('^')
            '^'

            >>> abjad.String.to_tridirectional_lilypond_symbol('-')
            '-'

            >>> abjad.String.to_tridirectional_lilypond_symbol('_')
            '_'

        ..  container:: example

            Returns none when ``argument`` is none:

            >>> abjad.String.to_tridirectional_lilypond_symbol(None) is None
            True

        """
        if argument is None:
            return None
        try:
            alignment = enums.VerticalAlignment.from_expr(argument)
        except Exception:
            raise ValueError(repr(argument))
        return String(format(alignment, 'lilypond'))

    @staticmethod
    def to_tridirectional_ordinal_constant(
        argument,
        ) -> typing.Union[None, enums.VerticalAlignment, 'String']:
        """
        Changes ``argument`` to tridirectional ordinal constant.

        ..  container:: example

            >>> abjad.String.to_tridirectional_ordinal_constant('^')
            Up

            >>> abjad.String.to_tridirectional_ordinal_constant('_')
            Down

            >>> abjad.String.to_tridirectional_ordinal_constant(1)
            Up

            >>> abjad.String.to_tridirectional_ordinal_constant(-1)
            Down

            >>> abjad.String.to_tridirectional_ordinal_constant(abjad.Up)
            Up

            >>> abjad.String.to_tridirectional_ordinal_constant(abjad.Down)
            Down

            >>> abjad.String.to_tridirectional_ordinal_constant(abjad.Center)
            Center

        ..  container:: example

            >>> abjad.String.to_tridirectional_ordinal_constant(None) is None
            True

        """
        if argument is None:
            return None
        return enums.VerticalAlignment.from_expr(argument)

    def to_upper_camel_case(self) -> 'String':
        """
        Changes string to upper camel case.

        ..  container:: example

            >>> abjad.String('scale degrees 4 and 5').to_upper_camel_case()
            'ScaleDegrees4And5'

            >>> abjad.String('scale_degrees_4_and_5').to_upper_camel_case()
            'ScaleDegrees4And5'

            >>> abjad.String('scale-degrees-4-and-5').to_upper_camel_case()
            'ScaleDegrees4And5'

            >>> abjad.String('ScaleDegrees4And5').to_upper_camel_case()
            'ScaleDegrees4And5'

        """
        words = self.delimit_words()
        words_ = [_.capitalize() for _ in words]
        string = ''.join(words_)
        return type(self)(string)



"""
Custom exceptions.
"""

class AssignabilityError(Exception):
    r'''Duration can not be assigned to note, rest or chord.
    '''
    pass


class ExtraSpannerError(Exception):
    r'''More than one spanner found for single-spanner operation.
    '''
    pass


class ImpreciseMetronomeMarkError(Exception):
    r'''MetronomeMark is imprecise.
    '''
    pass


class LilyPondParserError(Exception):
    r'''Can not parse input.
    '''
    pass


class MissingMeasureError(Exception):
    r'''No measure found.
    '''
    pass


class MissingMetronomeMarkError(Exception):
    r'''No metronome mark found.
    '''
    pass


class MissingSpannerError(Exception):
    r'''No spanner found.
    '''
    pass


class OverfullContainerError(Exception):
    r'''Container contents duration is greater than container target duration.
    '''
    pass


class ParentageError(Exception):
    r'''A parentage error.
    '''
    pass


class PersistentIndicatorError(Exception):
    r'''Persistent indicator already attached at same context.
    '''
    pass


class SchemeParserFinishedError(Exception):
    r'''SchemeParser has finished parsing.
    '''
    pass


class UnboundedTimeIntervalError(Exception):
    r'''Time interval has no bounds.
    '''
    pass


class UnderfullContainerError(Exception):
    r'''Container contents duration is less than container target duration.
    '''
    pass


class WellformednessError(Exception):
    r'''Score not well formed.
    '''
    pass



import typing
from abjad.utilities.String import String
from .AbjadValueObject import AbjadValueObject


class Tags(AbjadValueObject):
    """
    Tags.

    ..  container:: example

        Abjad-level singleton:

        >>> abjad.tags
        Tags()
        
    ..  container:: example

        Class is also available:

        >>> abjad.Tags()
        Tags()

    ..  container:: example

        >>> abjad.tags.SHOW_TO_JOIN_BROKEN_SPANNERS
        'SHOW_TO_JOIN_BROKEN_SPANNERS'

    """

    ### CLASS VARIABLES ###

    __slots__ = (
        )

    _known_tags: tuple = (

        ### BROKEN SPANNERS ###

        'HIDE_TO_JOIN_BROKEN_SPANNERS',
        'SHOW_TO_JOIN_BROKEN_SPANNERS',
        'RIGHT_BROKEN_BEAM', # used in figure-maker

        ### CLEFS ###

        'DEFAULT_CLEF',
        'DEFAULT_CLEF_COLOR',
        'DEFAULT_CLEF_COLOR_CANCELLATION',
        'DEFAULT_CLEF_REDRAW_COLOR',

        'EXPLICIT_CLEF',
        'EXPLICIT_CLEF_COLOR',
        'EXPLICIT_CLEF_COLOR_CANCELLATION',
        'EXPLICIT_CLEF_REDRAW_COLOR',

        'REAPPLIED_CLEF',
        'REAPPLIED_CLEF_COLOR',
        'REAPPLIED_CLEF_COLOR_CANCELLATION',
        'REAPPLIED_CLEF_REDRAW_COLOR',

        'REDUNDANT_CLEF',
        'REDUNDANT_CLEF_COLOR',
        'REDUNDANT_CLEF_COLOR_CANCELLATION',
        'REDUNDANT_CLEF_REDRAW_COLOR',

        ### DOCUMENT COMMANDS ###

        'BREAK',
        'EMPTY_START_BAR',
        'EOL_FERMATA',
        'FERMATA_MEASURE',
        'SHIFTED_CLEF',
        'SPACING',
        'SPACING_OVERRIDE',
        'TWO_VOICE',

        ### DOCUMENT MARKUP ###

        'CLOCK_TIME_MARKUP',
        'FIGURE_NAME_MARKUP',
        'LOCAL_MEASURE_NUMBER_MARKUP',
        'MEASURE_INDEX_MARKUP',
        'MEASURE_NUMBER_MARKUP',
        'SPACING_MARKUP',
        'SPACING_OVERRIDE_MARKUP',
        'STAGE_NUMBER_MARKUP',

        ### DOCUMENT TYPES ###

        'BUILD',
        'FIRST_SEGMENT_DEFAULT',
        'PARTS',
        'SCORE',
        'SEGMENT',

        ### DYNAMICS ###

        'EXPLICIT_DYNAMIC',
        'EXPLICIT_DYNAMIC_COLOR',
        'EXPLICIT_DYNAMIC_COLOR_CANCELLATION',
        'EXPLICIT_DYNAMIC_REDRAW_COLOR',

        'REAPPLIED_DYNAMIC',
        'REAPPLIED_DYNAMIC_COLOR',
        'REAPPLIED_DYNAMIC_COLOR_CANCELLATION',
        'REAPPLIED_DYNAMIC_REDRAW_COLOR',

        'REDUNDANT_DYNAMIC',
        'REDUNDANT_DYNAMIC_COLOR',
        'REDUNDANT_DYNAMIC_COLOR_CANCELLATION',
        'REDUNDANT_DYNAMIC_REDRAW_COLOR',

        ### FIGURES ###

        'FORESHADOW',
        'INCOMPLETE',
        'RECOLLECTION',

        ### INSTRUMENTS ###

        'DEFAULT_INSTRUMENT',
        'DEFAULT_INSTRUMENT_ALERT',
        'DEFAULT_INSTRUMENT_COLOR',
        'REDRAWN_DEFAULT_INSTRUMENT',
        'REDRAWN_DEFAULT_INSTRUMENT_COLOR',

        'EXPLICIT_INSTRUMENT',
        'EXPLICIT_INSTRUMENT_ALERT',
        'EXPLICIT_INSTRUMENT_COLOR',
        'REDRAWN_EXPLICIT_INSTRUMENT',
        'REDRAWN_EXPLICIT_INSTRUMENT_COLOR',

        'REAPPLIED_INSTRUMENT',
        'REAPPLIED_INSTRUMENT_ALERT',
        'REAPPLIED_INSTRUMENT_COLOR',
        'REDRAWN_REAPPLIED_INSTRUMENT',
        'REDRAWN_REAPPLIED_INSTRUMENT_COLOR',

        'REDUNDANT_INSTRUMENT',
        'REDUNDANT_INSTRUMENT_ALERT',
        'REDUNDANT_INSTRUMENT_COLOR',
        'REDRAWN_REDUNDANT_INSTRUMENT',
        'REDRAWN_REDUNDANT_INSTRUMENT_COLOR',

        ### MARGIN MARKUP ###

        'DEFAULT_MARGIN_MARKUP',
        'DEFAULT_MARGIN_MARKUP_ALERT',
        'DEFAULT_MARGIN_MARKUP_COLOR',
        'REDRAWN_DEFAULT_MARGIN_MARKUP',
        'REDRAWN_DEFAULT_MARGIN_MARKUP_COLOR',

        'EXPLICIT_MARGIN_MARKUP',
        'EXPLICIT_MARGIN_MARKUP_ALERT',
        'EXPLICIT_MARGIN_MARKUP_COLOR',
        'REDRAWN_EXPLICIT_MARGIN_MARKUP',
        'REDRAWN_EXPLICIT_MARGIN_MARKUP_COLOR',

        'REAPPLIED_MARGIN_MARKUP',
        'REAPPLIED_MARGIN_MARKUP_ALERT',
        'REAPPLIED_MARGIN_MARKUP_COLOR',
        'REDRAWN_REAPPLIED_MARGIN_MARKUP',
        'REDRAWN_REAPPLIED_MARGIN_MARKUP_COLOR',

        'REDUNDANT_MARGIN_MARKUP',
        'REDUNDANT_MARGIN_MARKUP_ALERT',
        'REDUNDANT_MARGIN_MARKUP_COLOR',
        'REDRAWN_REDUNDANT_MARGIN_MARKUP',
        'REDRAWN_REDUNDANT_MARGIN_MARKUP_COLOR',

        ### METRONOME MARKS ###

        'EXPLICIT_METRONOME_MARK',
        'EXPLICIT_METRONOME_MARK_WITH_COLOR',

        'REAPPLIED_METRONOME_MARK',
        'REAPPLIED_METRONOME_MARK_WITH_COLOR',

        'REDUNDANT_METRONOME_MARK',
        'REDUNDANT_METRONOME_MARK_WITH_COLOR',

        ### PERSISTENT OVERRIDES ###

        'EXPLICIT_PERSISTENT_OVERRIDE',

        'REAPPLIED_PERSISTENT_OVERRIDE',

        'REDUNDANT_PERSISTENT_OVERRIDE',

        ### PITCH HANDLING ###

        'ALLOW_OCTAVE',
        'ALLOW_OUT_OF_RANGE',
        'ALLOW_REPEAT_PITCH',
        'DO_NOT_TRANSPOSE',
        'NOT_YET_PITCHED',
        'NOT_YET_REGISTERED',
        'STAFF_POSITION',

        ### SPACING SECTION ###

        'EXPLICIT_SPACING_SECTION',
        'EXPLICIT_SPACING_SECTION_COLOR',

        'REAPPLIED_SPACING_SECTION',
        'REAPPLIED_SPACING_SECTION_COLOR',

        'REDUNDANT_SPACING_SECTION',
        'REDUNDANT_SPACING_SECTION_COLOR',

        ### STAFF LINES ###

        'EXPLICIT_STAFF_LINES',
        'EXPLICIT_STAFF_LINES_COLOR',

        'REAPPLIED_STAFF_LINES',
        'REAPPLIED_STAFF_LINES_COLOR',

        'REDUNDANT_STAFF_LINES',
        'REDUNDANT_STAFF_LINES_COLOR',

        ### TIME SIGNATURES ###

        'EXPLICIT_TIME_SIGNATURE',
        'EXPLICIT_TIME_SIGNATURE_COLOR',

        'REAPPLIED_TIME_SIGNATURE',
        'REAPPLIED_TIME_SIGNATURE_COLOR',

        'REDUNDANT_TIME_SIGNATURE',
        'REDUNDANT_TIME_SIGNATURE_COLOR',

        ### ZZZ: OTHER ###

        'LEFT_BROKEN_REPEAT_TIE_TO',
        'PITCH',
        'REMOVE_ALL_EMPTY_STAVES',
        'REPEAT_TIE',
        'RHYTHM',
        'RIGHT_BROKEN_TIE_FROM',
        'SOUNDS_DURING_SEGMENT',
        'TEMPORARY_CONTAINER',
        'TIE_FROM',
        'TIE_TO',

    )

    ### SPECIAL METHODS ###

    def __getattr__(self, tag: str) -> str:
        """
        Gets ``tag``.

        Raises attribute error when ``tag`` is unknown.
        """
        if tag not in self._known_tags:
            raise AttributeError('unknown tag {!r}.'.format(tag))
        return tag

    ### PUBLIC METHODS ###

    def clef_color_tags(self, path=None) -> typing.List[str]:
        """
        Gets clef color tags.

        ..  container:: example

            >>> for tag in abjad.tags.clef_color_tags():
            ...     tag
            ...
            'DEFAULT_CLEF_COLOR'
            'DEFAULT_CLEF_REDRAW_COLOR'
            'EXPLICIT_CLEF_COLOR'
            'EXPLICIT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF_COLOR'
            'REAPPLIED_CLEF_REDRAW_COLOR'
            'REDUNDANT_CLEF_COLOR'
            'REDUNDANT_CLEF_REDRAW_COLOR'

        ..  container:: example

            Segment:

            >>> path = abjad.Path('etude', 'segments', '_')
            >>> for tag in abjad.tags.clef_color_tags(path=path):
            ...     tag
            ...
            'DEFAULT_CLEF_COLOR'
            'DEFAULT_CLEF_REDRAW_COLOR'
            'EXPLICIT_CLEF_COLOR'
            'EXPLICIT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF_COLOR'
            'REAPPLIED_CLEF_REDRAW_COLOR'
            'REDUNDANT_CLEF_COLOR'
            'REDUNDANT_CLEF_REDRAW_COLOR'

        ..  container:: example

            Build:

            >>> path = abjad.Path('etude', 'builds', 'letter-score')
            >>> for tag in abjad.tags.clef_color_tags(path=path):
            ...     tag
            ...
            'DEFAULT_CLEF_COLOR'
            'DEFAULT_CLEF_REDRAW_COLOR'
            'EXPLICIT_CLEF_COLOR'
            'EXPLICIT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF_COLOR'
            'REAPPLIED_CLEF_REDRAW_COLOR'
            'REDUNDANT_CLEF_COLOR'
            'REDUNDANT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF'

        """
        tags = [
            self.DEFAULT_CLEF_COLOR,
            self.DEFAULT_CLEF_REDRAW_COLOR,
            self.EXPLICIT_CLEF_COLOR,
            self.EXPLICIT_CLEF_REDRAW_COLOR,
            self.REAPPLIED_CLEF_COLOR,
            self.REAPPLIED_CLEF_REDRAW_COLOR,
            self.REDUNDANT_CLEF_COLOR,
            self.REDUNDANT_CLEF_REDRAW_COLOR,
            ]
        if path and not path.is_segment():
            tags.append(self.REAPPLIED_CLEF)
        return tags

    def documentation_removal_tags(self):
        """
        Gets documentation removal tags.

        ..  container:: example

            >>> for tag in abjad.tags.documentation_removal_tags():
            ...     tag
            ...
            'CLOCK_TIME_MARKUP'
            'FIGURE_NAME_MARKUP'
            'LOCAL_MEASURE_NUMBER_MARKUP'
            'MEASURE_INDEX_MARKUP'
            'MEASURE_NUMBER_MARKUP'
            'SPACING_MARKUP'
            'STAGE_NUMBER_MARKUP'

        """
        return [
            self.CLOCK_TIME_MARKUP,
            self.FIGURE_NAME_MARKUP,
            self.LOCAL_MEASURE_NUMBER_MARKUP,
            self.MEASURE_INDEX_MARKUP,
            self.MEASURE_NUMBER_MARKUP,
            self.SPACING_MARKUP,
            self.STAGE_NUMBER_MARKUP,
            ]

    def dynamic_color_tags(self, path=None) -> typing.List[str]:
        """
        Gets dynamic color tags.

        ..  container:: example

            >>> for tag in abjad.tags.dynamic_color_tags():
            ...     tag
            ...
            'EXPLICIT_DYNAMIC_COLOR'
            'EXPLICIT_DYNAMIC_REDRAW_COLOR'
            'REAPPLIED_DYNAMIC'
            'REAPPLIED_DYNAMIC_COLOR'
            'REAPPLIED_DYNAMIC_REDRAW_COLOR'
            'REDUNDANT_DYNAMIC_COLOR'
            'REDUNDANT_DYNAMIC_REDRAW_COLOR'

        Ignores ``path``.
        """
        return [
            self.EXPLICIT_DYNAMIC_COLOR,
            self.EXPLICIT_DYNAMIC_REDRAW_COLOR,
            self.REAPPLIED_DYNAMIC,
            self.REAPPLIED_DYNAMIC_COLOR,
            self.REAPPLIED_DYNAMIC_REDRAW_COLOR,
            self.REDUNDANT_DYNAMIC_COLOR,
            self.REDUNDANT_DYNAMIC_REDRAW_COLOR,
            ]

    def instrument_color_tags(self, path=None) -> typing.List[str]:
        """
        Gets instrument color tags.

        ..  container:: example

            >>> for tag in abjad.tags.instrument_color_tags():
            ...     tag
            ...
            'DEFAULT_INSTRUMENT_ALERT'
            'DEFAULT_INSTRUMENT_COLOR'
            'REDRAWN_DEFAULT_INSTRUMENT_COLOR'
            'EXPLICIT_INSTRUMENT_ALERT'
            'EXPLICIT_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_ALERT'
            'REDRAWN_EXPLICIT_INSTRUMENT_COLOR'
            'REDRAWN_REAPPLIED_INSTRUMENT_COLOR'
            'REDUNDANT_INSTRUMENT_ALERT'
            'REDUNDANT_INSTRUMENT_COLOR'
            'REDRAWN_REDUNDANT_INSTRUMENT_COLOR'

        Ignores ``path``.
        """
        return [
            self.DEFAULT_INSTRUMENT_ALERT,
            self.DEFAULT_INSTRUMENT_COLOR,
            self.REDRAWN_DEFAULT_INSTRUMENT_COLOR,
            self.EXPLICIT_INSTRUMENT_ALERT,
            self.EXPLICIT_INSTRUMENT_COLOR,
            self.REAPPLIED_INSTRUMENT_COLOR,
            self.REAPPLIED_INSTRUMENT_ALERT,
            self.REDRAWN_EXPLICIT_INSTRUMENT_COLOR,
            self.REDRAWN_REAPPLIED_INSTRUMENT_COLOR,
            self.REDUNDANT_INSTRUMENT_ALERT,
            self.REDUNDANT_INSTRUMENT_COLOR,
            self.REDRAWN_REDUNDANT_INSTRUMENT_COLOR,
            ]

    def layout_removal_tags(self):
        """
        Gets layout removal tags.

        ..  container:: example

            >>> for tag in abjad.tags.layout_removal_tags():
            ...     tag
            ...
            'EMPTY_START_BAR'
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'LOCAL_MEASURE_NUMBER_MARKUP'
            'MEASURE_INDEX_MARKUP'
            'MEASURE_NUMBER_MARKUP'
            'REDUNDANT_TIME_SIGNATURE_COLOR'
            'STAGE_NUMBER_MARKUP'

        """
        return [
            self.EMPTY_START_BAR,
            self.EXPLICIT_TIME_SIGNATURE_COLOR,
            self.LOCAL_MEASURE_NUMBER_MARKUP,
            self.MEASURE_INDEX_MARKUP,
            self.MEASURE_NUMBER_MARKUP,
            self.REDUNDANT_TIME_SIGNATURE_COLOR,
            self.STAGE_NUMBER_MARKUP,
            ]

    def margin_markup_color_tags(self, path=None) -> typing.List[str]:
        """
        Gets margin markup color tags.

        ..  container:: example

            >>> for tag in abjad.tags.margin_markup_color_tags():
            ...     tag
            ...
            'DEFAULT_MARGIN_MARKUP_ALERT'
            'DEFAULT_MARGIN_MARKUP_COLOR'
            'REDRAWN_DEFAULT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_MARGIN_MARKUP_ALERT'
            'EXPLICIT_MARGIN_MARKUP_COLOR'
            'REAPPLIED_MARGIN_MARKUP_ALERT'
            'REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDRAWN_EXPLICIT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDUNDANT_MARGIN_MARKUP_ALERT'
            'REDUNDANT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REDUNDANT_MARGIN_MARKUP_COLOR'

        Ignores ``path``.
        """
        return [
            self.DEFAULT_MARGIN_MARKUP_ALERT,
            self.DEFAULT_MARGIN_MARKUP_COLOR,
            self.REDRAWN_DEFAULT_MARGIN_MARKUP_COLOR,
            self.EXPLICIT_MARGIN_MARKUP_ALERT,
            self.EXPLICIT_MARGIN_MARKUP_COLOR,
            self.REAPPLIED_MARGIN_MARKUP_ALERT,
            self.REAPPLIED_MARGIN_MARKUP_COLOR,
            self.REDRAWN_EXPLICIT_MARGIN_MARKUP_COLOR,
            self.REDRAWN_REAPPLIED_MARGIN_MARKUP_COLOR,
            self.REDUNDANT_MARGIN_MARKUP_ALERT,
            self.REDUNDANT_MARGIN_MARKUP_COLOR,
            self.REDRAWN_REDUNDANT_MARGIN_MARKUP_COLOR,
            ]

    def metronome_mark_color_expression_tags(self, path=None) -> typing.List[str]:
        """
        Gets metronome mark color expression tags.

        ..  container:: example

            >>> for tag in abjad.tags.metronome_mark_color_expression_tags():
            ...     tag
            ...
            'EXPLICIT_METRONOME_MARK_WITH_COLOR'
            'REAPPLIED_METRONOME_MARK_WITH_COLOR'
            'REDUNDANT_METRONOME_MARK_WITH_COLOR'

        """
        return [
            self.EXPLICIT_METRONOME_MARK_WITH_COLOR,
            self.REAPPLIED_METRONOME_MARK_WITH_COLOR,
            self.REDUNDANT_METRONOME_MARK_WITH_COLOR,
            ]

    def metronome_mark_color_suppression_tags(self, path=None) -> typing.List[str]:
        """
        Gets metronome mark color suppression tags.

        ..  container:: example

            >>> for tag in abjad.tags.metronome_mark_color_suppression_tags():
            ...     tag
            ...
            'EXPLICIT_METRONOME_MARK'
            'REDUNDANT_METRONOME_MARK'

        Ignores ``path``.
        """
        return [
            self.EXPLICIT_METRONOME_MARK,
            self.REDUNDANT_METRONOME_MARK,
            ]

    def music_annotation_tags(self) -> typing.List[str]:
        """
        Gets music annotation tags.

        ..  container:: example

            >>> for tag in abjad.tags.music_annotation_tags():
            ...     tag
            ...
            'CLOCK_TIME_MARKUP'
            'FIGURE_NAME_MARKUP'
            'MEASURE_NUMBER_MARKUP'
            'SPACING_MARKUP'
            'SPACING_OVERRIDE_MARKUP'
            'STAGE_NUMBER_MARKUP'

        """
        return [
            self.CLOCK_TIME_MARKUP,
            self.FIGURE_NAME_MARKUP,
            self.MEASURE_NUMBER_MARKUP,
            self.SPACING_MARKUP,
            self.SPACING_OVERRIDE_MARKUP,
            self.STAGE_NUMBER_MARKUP,
            ]

    def persistent_indicator_color_expression_tags(
        self,
        path=None,
        ) -> typing.List[str]:
        """
        Gets persistent indicator color expression tags.

        ..  container:: example

            >>> tags = abjad.tags.persistent_indicator_color_expression_tags()
            >>> for tag in tags:
            ...     tag
            ...
            'DEFAULT_CLEF_COLOR'
            'DEFAULT_CLEF_REDRAW_COLOR'
            'EXPLICIT_CLEF_COLOR'
            'EXPLICIT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF_COLOR'
            'REAPPLIED_CLEF_REDRAW_COLOR'
            'REDUNDANT_CLEF_COLOR'
            'REDUNDANT_CLEF_REDRAW_COLOR'
            'EXPLICIT_DYNAMIC_COLOR'
            'EXPLICIT_DYNAMIC_REDRAW_COLOR'
            'REAPPLIED_DYNAMIC'
            'REAPPLIED_DYNAMIC_COLOR'
            'REAPPLIED_DYNAMIC_REDRAW_COLOR'
            'REDUNDANT_DYNAMIC_COLOR'
            'REDUNDANT_DYNAMIC_REDRAW_COLOR'
            'DEFAULT_INSTRUMENT_ALERT'
            'DEFAULT_INSTRUMENT_COLOR'
            'REDRAWN_DEFAULT_INSTRUMENT_COLOR'
            'EXPLICIT_INSTRUMENT_ALERT'
            'EXPLICIT_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_ALERT'
            'REDRAWN_EXPLICIT_INSTRUMENT_COLOR'
            'REDRAWN_REAPPLIED_INSTRUMENT_COLOR'
            'REDUNDANT_INSTRUMENT_ALERT'
            'REDUNDANT_INSTRUMENT_COLOR'
            'REDRAWN_REDUNDANT_INSTRUMENT_COLOR'
            'DEFAULT_MARGIN_MARKUP_ALERT'
            'DEFAULT_MARGIN_MARKUP_COLOR'
            'REDRAWN_DEFAULT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_MARGIN_MARKUP_ALERT'
            'EXPLICIT_MARGIN_MARKUP_COLOR'
            'REAPPLIED_MARGIN_MARKUP_ALERT'
            'REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDRAWN_EXPLICIT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDUNDANT_MARGIN_MARKUP_ALERT'
            'REDUNDANT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REDUNDANT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_METRONOME_MARK_WITH_COLOR'
            'REAPPLIED_METRONOME_MARK_WITH_COLOR'
            'REDUNDANT_METRONOME_MARK_WITH_COLOR'
            'EXPLICIT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES_COLOR'
            'REDUNDANT_STAFF_LINES_COLOR'
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE_COLOR'
            'REDUNDANT_TIME_SIGNATURE_COLOR'

        ..  container:: example

            Segment:

            >>> path = abjad.Path('etude', 'segments', '_')
            >>> tags = abjad.tags.persistent_indicator_color_expression_tags(path)
            >>> for tag in tags:
            ...     tag
            ...
            'DEFAULT_CLEF_COLOR'
            'DEFAULT_CLEF_REDRAW_COLOR'
            'EXPLICIT_CLEF_COLOR'
            'EXPLICIT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF_COLOR'
            'REAPPLIED_CLEF_REDRAW_COLOR'
            'REDUNDANT_CLEF_COLOR'
            'REDUNDANT_CLEF_REDRAW_COLOR'
            'EXPLICIT_DYNAMIC_COLOR'
            'EXPLICIT_DYNAMIC_REDRAW_COLOR'
            'REAPPLIED_DYNAMIC'
            'REAPPLIED_DYNAMIC_COLOR'
            'REAPPLIED_DYNAMIC_REDRAW_COLOR'
            'REDUNDANT_DYNAMIC_COLOR'
            'REDUNDANT_DYNAMIC_REDRAW_COLOR'
            'DEFAULT_INSTRUMENT_ALERT'
            'DEFAULT_INSTRUMENT_COLOR'
            'REDRAWN_DEFAULT_INSTRUMENT_COLOR'
            'EXPLICIT_INSTRUMENT_ALERT'
            'EXPLICIT_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_ALERT'
            'REDRAWN_EXPLICIT_INSTRUMENT_COLOR'
            'REDRAWN_REAPPLIED_INSTRUMENT_COLOR'
            'REDUNDANT_INSTRUMENT_ALERT'
            'REDUNDANT_INSTRUMENT_COLOR'
            'REDRAWN_REDUNDANT_INSTRUMENT_COLOR'
            'DEFAULT_MARGIN_MARKUP_ALERT'
            'DEFAULT_MARGIN_MARKUP_COLOR'
            'REDRAWN_DEFAULT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_MARGIN_MARKUP_ALERT'
            'EXPLICIT_MARGIN_MARKUP_COLOR'
            'REAPPLIED_MARGIN_MARKUP_ALERT'
            'REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDRAWN_EXPLICIT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDUNDANT_MARGIN_MARKUP_ALERT'
            'REDUNDANT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REDUNDANT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_METRONOME_MARK_WITH_COLOR'
            'REAPPLIED_METRONOME_MARK_WITH_COLOR'
            'REDUNDANT_METRONOME_MARK_WITH_COLOR'
            'EXPLICIT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES_COLOR'
            'REDUNDANT_STAFF_LINES_COLOR'
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE_COLOR'
            'REDUNDANT_TIME_SIGNATURE_COLOR'

        ..  container:: example

            Build:

            >>> path = abjad.Path('etude', 'builds', 'letter-score')
            >>> tags = abjad.tags.persistent_indicator_color_expression_tags(path)
            >>> for tag in tags:
            ...     tag
            ...
            'DEFAULT_CLEF_COLOR'
            'DEFAULT_CLEF_REDRAW_COLOR'
            'EXPLICIT_CLEF_COLOR'
            'EXPLICIT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF_COLOR'
            'REAPPLIED_CLEF_REDRAW_COLOR'
            'REDUNDANT_CLEF_COLOR'
            'REDUNDANT_CLEF_REDRAW_COLOR'
            'REAPPLIED_CLEF'
            'EXPLICIT_DYNAMIC_COLOR'
            'EXPLICIT_DYNAMIC_REDRAW_COLOR'
            'REAPPLIED_DYNAMIC'
            'REAPPLIED_DYNAMIC_COLOR'
            'REAPPLIED_DYNAMIC_REDRAW_COLOR'
            'REDUNDANT_DYNAMIC_COLOR'
            'REDUNDANT_DYNAMIC_REDRAW_COLOR'
            'DEFAULT_INSTRUMENT_ALERT'
            'DEFAULT_INSTRUMENT_COLOR'
            'REDRAWN_DEFAULT_INSTRUMENT_COLOR'
            'EXPLICIT_INSTRUMENT_ALERT'
            'EXPLICIT_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_COLOR'
            'REAPPLIED_INSTRUMENT_ALERT'
            'REDRAWN_EXPLICIT_INSTRUMENT_COLOR'
            'REDRAWN_REAPPLIED_INSTRUMENT_COLOR'
            'REDUNDANT_INSTRUMENT_ALERT'
            'REDUNDANT_INSTRUMENT_COLOR'
            'REDRAWN_REDUNDANT_INSTRUMENT_COLOR'
            'DEFAULT_MARGIN_MARKUP_ALERT'
            'DEFAULT_MARGIN_MARKUP_COLOR'
            'REDRAWN_DEFAULT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_MARGIN_MARKUP_ALERT'
            'EXPLICIT_MARGIN_MARKUP_COLOR'
            'REAPPLIED_MARGIN_MARKUP_ALERT'
            'REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDRAWN_EXPLICIT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REAPPLIED_MARGIN_MARKUP_COLOR'
            'REDUNDANT_MARGIN_MARKUP_ALERT'
            'REDUNDANT_MARGIN_MARKUP_COLOR'
            'REDRAWN_REDUNDANT_MARGIN_MARKUP_COLOR'
            'EXPLICIT_METRONOME_MARK_WITH_COLOR'
            'REAPPLIED_METRONOME_MARK_WITH_COLOR'
            'REDUNDANT_METRONOME_MARK_WITH_COLOR'
            'EXPLICIT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES_COLOR'
            'REDUNDANT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES'
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE_COLOR'
            'REDUNDANT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE'

        """
        tags: typing.List[str] = []
        tags.extend(self.clef_color_tags(path))
        tags.extend(self.dynamic_color_tags(path))
        tags.extend(self.instrument_color_tags(path))
        tags.extend(self.margin_markup_color_tags(path))
        tags.extend(self.metronome_mark_color_expression_tags(path))
        tags.extend(self.staff_lines_color_tags(path))
        tags.extend(self.time_signature_color_tags(path))
        return tags

    def persistent_indicator_color_suppression_tags(
        self,
        path=None,
        ) -> typing.List[str]:
        """
        Gets persistent indicator color suppression tags.

        ..  container:: example

            >>> tags = abjad.tags.persistent_indicator_color_suppression_tags()
            >>> for tag in tags:
            ...     tag
            ...
            'EXPLICIT_METRONOME_MARK'
            'REDUNDANT_METRONOME_MARK'

        ..  container:: example

            Segment:

            >>> path = abjad.Path('etude', 'segments', '_')
            >>> tags = abjad.tags.persistent_indicator_color_suppression_tags(path)
            >>> for tag in tags:
            ...     tag
            ...
            'EXPLICIT_METRONOME_MARK'
            'REDUNDANT_METRONOME_MARK'

        ..  container:: example

            Build:

            >>> path = abjad.Path('etude', 'builds', 'letter-score')
            >>> tags = abjad.tags.persistent_indicator_color_suppression_tags(path)
            >>> for tag in tags:
            ...     tag
            ...
            'EXPLICIT_METRONOME_MARK'
            'REDUNDANT_METRONOME_MARK'

        """
        tags: typing.List[str] = []
        tags.extend(self.metronome_mark_color_suppression_tags())
        return tags

    def persistent_indicator_tags(self) -> typing.List[str]:
        """
        Gets persistence tags.

        ..  container:: example

            >>> for string in abjad.tags.persistent_indicator_tags():
            ...     string
            ...
            'DEFAULT_CLEF'
            'EXPLICIT_CLEF'
            'REAPPLIED_CLEF'
            'REDUNDANT_CLEF'
            'EXPLICIT_DYNAMIC'
            'REAPPLIED_DYNAMIC'
            'REDUNDANT_DYNAMIC'
            'DEFAULT_INSTRUMENT'
            'EXPLICIT_INSTRUMENT'
            'REAPPLIED_INSTRUMENT'
            'REDUNDANT_INSTRUMENT'
            'DEFAULT_MARGIN_MARKUP'
            'EXPLICIT_MARGIN_MARKUP'
            'REAPPLIED_MARGIN_MARKUP'
            'REDUNDANT_MARGIN_MARKUP'
            'EXPLICIT_METRONOME_MARK'
            'REAPPLIED_METRONOME_MARK'
            'REDUNDANT_METRONOME_MARK'
            'EXPLICIT_PERSISTENT_OVERRIDE'
            'REAPPLIED_PERSISTENT_OVERRIDE'
            'REDUNDANT_PERSISTENT_OVERRIDE'
            'EXPLICIT_STAFF_LINES'
            'REAPPLIED_STAFF_LINES'
            'REDUNDANT_STAFF_LINES'
            'EXPLICIT_TIME_SIGNATURE'
            'REAPPLIED_TIME_SIGNATURE'
            'REDUNDANT_TIME_SIGNATURE'

        """
        return [
            self.DEFAULT_CLEF,
            self.EXPLICIT_CLEF,
            self.REAPPLIED_CLEF,
            self.REDUNDANT_CLEF,
            #
            self.EXPLICIT_DYNAMIC,
            self.REAPPLIED_DYNAMIC,
            self.REDUNDANT_DYNAMIC,
            #
            self.DEFAULT_INSTRUMENT,
            self.EXPLICIT_INSTRUMENT,
            self.REAPPLIED_INSTRUMENT,
            self.REDUNDANT_INSTRUMENT,
            #
            self.DEFAULT_MARGIN_MARKUP,
            self.EXPLICIT_MARGIN_MARKUP,
            self.REAPPLIED_MARGIN_MARKUP,
            self.REDUNDANT_MARGIN_MARKUP,
            #
            self.EXPLICIT_METRONOME_MARK,
            self.REAPPLIED_METRONOME_MARK,
            self.REDUNDANT_METRONOME_MARK,
            #
            self.EXPLICIT_PERSISTENT_OVERRIDE,
            self.REAPPLIED_PERSISTENT_OVERRIDE,
            self.REDUNDANT_PERSISTENT_OVERRIDE,
            #
            self.EXPLICIT_STAFF_LINES,
            self.REAPPLIED_STAFF_LINES,
            self.REDUNDANT_STAFF_LINES,
            #
            self.EXPLICIT_TIME_SIGNATURE,
            self.REAPPLIED_TIME_SIGNATURE,
            self.REDUNDANT_TIME_SIGNATURE,
            #
            ]

    def spacing_markup_tags(self) -> typing.List[str]:
        """
        Gets markup spacing tags.

        ..  container:: example

            >>> for tag in abjad.tags.spacing_markup_tags():
            ...     tag
            ...
            'SPACING_MARKUP'
            'SPACING_OVERRIDE_MARKUP'

        """
        return [
            self.SPACING_MARKUP,
            self.SPACING_OVERRIDE_MARKUP,
            ]

    def spacing_tags(self) -> typing.List[str]:
        """
        Gets spacing tags.

        ..  container:: example

            >>> for tag in abjad.tags.spacing_tags():
            ...     tag
            ...
            'SPACING'
            'SPACING_MARKUP'
            'SPACING_OVERRIDE'
            'SPACING_OVERRIDE_MARKUP'

        """
        return [
            self.SPACING,
            self.SPACING_MARKUP,
            self.SPACING_OVERRIDE,
            self.SPACING_OVERRIDE_MARKUP,
            ]

    def staff_lines_color_tags(self, path=None) -> typing.List[str]:
        """
        Gets staff lines color tags.

        ..  container:: example

            >>> for tag in abjad.tags.staff_lines_color_tags():
            ...     tag
            ...
            'EXPLICIT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES_COLOR'
            'REDUNDANT_STAFF_LINES_COLOR'

        ..  container:: example

            Segment:

            >>> path = abjad.Path('etude', 'segments', '_')
            >>> for tag in abjad.tags.staff_lines_color_tags(path):
            ...     tag
            ...
            'EXPLICIT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES_COLOR'
            'REDUNDANT_STAFF_LINES_COLOR'

        ..  container:: example

            Build:

            >>> path = abjad.Path('etude', 'builds', 'letter-score')
            >>> for tag in abjad.tags.staff_lines_color_tags(path):
            ...     tag
            ...
            'EXPLICIT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES_COLOR'
            'REDUNDANT_STAFF_LINES_COLOR'
            'REAPPLIED_STAFF_LINES'

        """
        tags = [
            self.EXPLICIT_STAFF_LINES_COLOR,
            self.REAPPLIED_STAFF_LINES_COLOR,
            self.REDUNDANT_STAFF_LINES_COLOR,
            ]
        if path and not path.is_segment():
            tags.append(self.REAPPLIED_STAFF_LINES)
        return tags

    def time_signature_color_tags(self, path=None) -> typing.List[str]:
        """
        Gets time signature color tags.

        ..  container:: example

            >>> for tag in abjad.tags.time_signature_color_tags():
            ...     tag
            ...
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE_COLOR'
            'REDUNDANT_TIME_SIGNATURE_COLOR'

        ..  container:: example

            Segment:

            >>> path = abjad.Path('etude', 'segments', '_')
            >>> for tag in abjad.tags.time_signature_color_tags():
            ...     tag
            ...
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE_COLOR'
            'REDUNDANT_TIME_SIGNATURE_COLOR'

        ..  container:: example

            Build:

            >>> path = abjad.Path('etude', 'builds', 'letter-score')
            >>> for tag in abjad.tags.time_signature_color_tags():
            ...     tag
            ...
            'EXPLICIT_TIME_SIGNATURE_COLOR'
            'REAPPLIED_TIME_SIGNATURE_COLOR'
            'REDUNDANT_TIME_SIGNATURE_COLOR'

        """
        tags = [
            self.EXPLICIT_TIME_SIGNATURE_COLOR,
            self.REAPPLIED_TIME_SIGNATURE_COLOR,
            self.REDUNDANT_TIME_SIGNATURE_COLOR,
            ]
        if path and not path.is_segment():
            tags.append(self.REAPPLIED_TIME_SIGNATURE)
        return tags



from .Leaf import Leaf


class Skip(Leaf):
    """
    LilyPond skip.

    ..  container:: example

        >>> skip = abjad.Skip((3, 16))
        >>> skip
        Skip('s8.')

        ..  docs::

            >>> abjad.f(skip)
            s8.

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Leaves'

    __slots__ = ()

    ### INITIALIZER ###

    def __init__(self, *arguments):
        import abjad
        input_leaf = None
        written_duration = None
        if len(arguments) == 1 and isinstance(arguments[0], str):
            string = f'{{ {arguments[0]} }}'
            parsed = abjad.parse(string)
            assert len(parsed) == 1 and isinstance(parsed[0], Leaf)
            input_leaf = parsed[0]
            written_duration = input_leaf.written_duration
        elif len(arguments) == 1 and isinstance(arguments[0], Leaf):
            written_duration = arguments[0].written_duration
            input_leaf = arguments[0]
        elif len(arguments) == 1 and not isinstance(arguments[0], str):
            written_duration = arguments[0]
        elif len(arguments) == 0:
            written_duration = abjad.Duration(1, 4)
        else:
            message = f'can not initialize skip from {arguments!r}.'
            raise ValueError(message)
        Leaf.__init__(self, written_duration)
        if input_leaf is not None:
            self._copy_override_and_set_from_leaf(input_leaf)

    ### PRIVATE METHODS ###

    def _get_body(self):
        result = []
        result.append(f's{self._get_formatted_duration()}')
        return result

    def _get_compact_representation(self):
        return f's{self._get_formatted_duration()}'



"""
Literature examples implemented in Abjad.
"""

from abjad.demos import bartok
from abjad.demos import ferneyhough
from abjad.demos import ligeti
from abjad.demos import mozart
from abjad.demos import part



from abjad.utilities.TypedList import TypedList


class CompoundInequality(TypedList):
    """
    Compound time-relation inequality.

    ..  container:: example

        >>> compound_inequality = abjad.timespans.CompoundInequality([
        ...    abjad.timespans.CompoundInequality([
        ...        'timespan_1.start_offset <= timespan_2.start_offset',
        ...        'timespan_2.start_offset < timespan_1.stop_offset'],
        ...         logical_operator='and'),
        ...    abjad.timespans.CompoundInequality([
        ...        'timespan_2.start_offset <= timespan_1.start_offset',
        ...        'timespan_1.start_offset < timespan_2.stop_offset'],
        ...        logical_operator='and')],
        ...    logical_operator='or',
        ...    )

        >>> abjad.f(compound_inequality)
        abjad.timespans.CompoundInequality(
            [
                abjad.timespans.CompoundInequality(
                    [
                        abjad.TimespanInequality('timespan_1.start_offset <= timespan_2.start_offset'),
                        abjad.TimespanInequality('timespan_2.start_offset < timespan_1.stop_offset'),
                        ],
                    logical_operator='and',
                    ),
                abjad.timespans.CompoundInequality(
                    [
                        abjad.TimespanInequality('timespan_2.start_offset <= timespan_1.start_offset'),
                        abjad.TimespanInequality('timespan_1.start_offset < timespan_2.stop_offset'),
                        ],
                    logical_operator='and',
                    ),
                ],
            logical_operator='or',
            )

    """

    ### CLASS VARIABLES ###

    __documentation_section__ = 'Time relations'

    __slots__ = (
        '_logical_operator',
        '_name',
        )

    logical_operator_dictionary = {
        'and': '&',
        'or': '|',
        'xor': '^',
        }

    ### INITIALIZER ###

    def __init__(self, items=None, logical_operator='and'):
        TypedList.__init__(self, items=items)
        self._logical_operator = logical_operator

    ### PUBLIC METHODS ###

    def evaluate(
        self,
        timespan_1_start_offset,
        timespan_1_stop_offset,
        timespan_2_start_offset,
        timespan_2_stop_offset,
        ):
        """
        Evalutes compound inequality.

        Returns true or false.
        """
        from abjad import timespans
        truth_values = []
        for inequality in self:
            # TODO: compress the following two branches
            if isinstance(inequality, timespans.TimespanInequality):
                truth_value = inequality.evaluate(
                    timespan_1_start_offset, timespan_1_stop_offset,
                    timespan_2_start_offset, timespan_2_stop_offset)
                truth_values.append(truth_value)
            elif isinstance(inequality, type(self)):
                truth_value = inequality.evaluate(
                    timespan_1_start_offset, timespan_1_stop_offset,
                    timespan_2_start_offset, timespan_2_stop_offset)
                truth_values.append(truth_value)
        if self.logical_operator == 'and':
            truth_value = all(truth_values)
        elif self.logical_operator == 'or':
            truth_value = any(truth_values)
        elif self.logical_operator == 'xor':
            truth_value = bool(len([x for x in truth_values if x]) == 1)
        else:
            message = 'unknown logical operator: {!r}.'
            message = message.format(self.logical_operator)
            raise ValueError(message)
        return truth_value

    def evaluate_offset_inequality(
        self,
        timespan_start,
        timespan_stop,
        offset,
        ):
        """
        Evalutes offset inequality.

        Returns true or false.
        """
        from abjad import timespans
        truth_values = []
        for inequality in self:
            if isinstance(inequality, timespans.TimespanInequality):
                truth_value = inequality.evaluate_offset_inequality(
                    timespan_start, timespan_stop, offset)
                truth_values.append(truth_value)
            elif isinstance(inequality, type(self)):
                truth_value = inequality.evaluate_offset_inequality(
                    timespan_start, timespan_stop, offset)
                truth_values.append(truth_value)
            else:
                message = 'unknown inequality: {!r}.'
                message = message.format(inequality)
                raise TypeError(message)
        assert truth_values, repr(truth_values)
        if self.logical_operator == 'and':
            truth_value = all(truth_values)
        elif self.logical_operator == 'or':
            truth_value = any(truth_values)
        elif self.logical_operator == 'xor':
            truth_value = bool(len([x for x in truth_values if x]) == 1)
        else:
            message = 'unknown logical operator: {!r}.'
            message = message.format(self.logical_operator)
            raise ValueError(message)
        return truth_value

    def get_offset_indices(
        self,
        timespan_1,
        timespan_2_start_offsets,
        timespan_2_stop_offsets,
        ):
        """
        Gets offset indices of compound inequality.
        """
        from abjad import timespans as abjad_timespans
        timespans = abjad_timespans.TimespanList()
        for element in self:
            # TODO: compress the following two branches
            if isinstance(element, type(self)):
                result = element.get_offset_indices(
                    timespan_1,
                    timespan_2_start_offsets,
                    timespan_2_stop_offsets)
                timespans.extend(result)
            elif isinstance(element, abjad_timespans.TimespanInequality):
                offset_indices = element.get_offset_indices(
                    timespan_1,
                    timespan_2_start_offsets,
                    timespan_2_stop_offsets)
                timespan = abjad_timespans.Timespan(*offset_indices)
                timespans.append(timespan)
            else:
                message = 'unknown inequality: {!r}.'
                message = message(element)
                raise TypeError(message)
        if self.logical_operator == 'and':
            result = timespans.compute_logical_and()
        elif self.logical_operator == 'or':
            timespans.sort()
            result = timespans.compute_logical_or()
        elif self.logical_operator == 'xor':
            result = timespans.compute_logical_xor()
        else:
            message = 'unknown logical operator: {!r}.'
            message = message.format(self.logical_operator)
            raise ValueError(message)
        return result

    ### PRIVATE PROPERTIES ###

    @property
    def _item_coercer(self):
        from abjad import timespans as abjad_timespans
        def coerce_(argument):
            if isinstance(argument, str):
                return abjad_timespans.TimespanInequality(argument)
            elif isinstance(argument, abjad_timespans.TimespanInequality):
                return argument
            elif isinstance(argument, abjad_timespans.CompoundInequality):
                return argument
            else:
                raise TypeError(argument)
        return coerce_

    ### PUBLIC PROPERTIES ###

    @property
    def logical_operator(self):
        """
        Compound inequality logical operator.
        """
        return self._logical_operator
